package linprog;
import graph.Face;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * Adds functionality to MetricGraph.
 * Some technical properties of MetricGraphs used in Chapter 5 of "A second generation proof"
 * No constructor, all methods are static.
 */

public class GraphStructure {

  private GraphStructure() {
  }

  public static boolean isPureTriangle(MetricGraph G, graph.Vertex V,int arity) {
    if (V.size()!=arity) return false;
    graph.Face F = V.getAny();
    for (int i=0;i<V.size();i++) {
      if (V.next(F,i).size()!=3)
        return false;
      if (!G.isQRTET(V.next(F,i)))
        return false;
    }
    return true;
  }



  /**
   * returns a list of all verticex of pure triangle type (arity,0)
   */
  private static graph.Vertex[] getPureTriangle(MetricGraph G,int arity) {
    ArrayList A = new ArrayList();
    for (Enumeration E = G.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex) E.nextElement();
      if (isPureTriangle(G,V,arity)) {
        A.add(V);
      }
    }
    return (graph.Vertex[]) A.toArray(new graph.Vertex[0]);
  }

  /**
   * returns a list of all length 2 vertex-strings of pure triangles arity 5.
   */
  private static ArrayList get2String(MetricGraph G) {
    ArrayList A = new ArrayList();
    graph.Vertex[] Varray = getPureTriangle(G,5);
    for (int i=0;i<Varray.length;i++) {
      graph.Vertex V = Varray[i];
      graph.Face F = V.getAny();
      util.Eiffel.assert(V.size()==5);
      for (int j=0;j<5;j++) {
        graph.Vertex W = V.next(F,j).next(V,1);
        if (isPureTriangle(G,W,5))
          A.add(new graph.Vertex[] {V,W});
      }
    }
    return A;
  }

  /**
   * returns a list of all pure-triangle strings of length at most 4, arity 5.
   * each return object is graph.Vertex[];
   */
  public static Object[] getPureTriangleStrings(MetricGraph G) {
    ArrayList A = new ArrayList();
    graph.Vertex[] Varray = getPureTriangle(G,5);
    for (int i=0;i<Varray.length;i++) {
      A.add(new graph.Vertex[] {Varray[i]});
    }
    A.addAll(get2String(G));
    A.addAll(get3String(G));
    A.addAll(get4String(G));
    //verify results:
    for (Iterator I = A.iterator();I.hasNext();/*--*/) {
      graph.Vertex[] gV = (graph.Vertex[]) I.next();
      for (int i=0;i<gV.length;i++) {
        util.Eiffel.assert(isPureTriangle(G,gV[i],5));
      }
    }
    return A.toArray();
  }

  /**
   * returns a list of 3-strings of pure triangle type (5,0)
   * A 3-string {W0,V,W1} with W0 and W1 distinct and adjacent to V, but not adjacent to each other.
   * each return object is graph.Vertex[3].
   */
  private static ArrayList get3String(MetricGraph G) {
    ArrayList A = new ArrayList();
    int[] data0 = {0,1,2,3,4};
    int[] data1 = {2,3,4,0,1}; // pairs {i,j} of non-adjacent positions in a pentagon.
    graph.Vertex[] Varray = getPureTriangle(G,5);
    for (int i=0;i<Varray.length;i++) {
      graph.Vertex V = Varray[i];
      graph.Face F = V.getAny();
      util.Eiffel.assert(V.size()==5);
      for (int j=0;j<5;j++) {
        graph.Vertex W0 = V.next(F,data0[j]).next(V,1);
        graph.Vertex W1 = V.next(F,data1[j]).next(V,1);
        if (isPureTriangle(G,W0,5) && isPureTriangle(G,W1,5))
          A.add(new graph.Vertex[] {W0,V,W1});
      }
    }
    return A;
  }

  /**
   * returns interconnected clusters of 4 vertices of type (5,0).
   * each return object is graph.Vertex[4];
   */
  private static ArrayList get4String(MetricGraph G) {
    ArrayList A = new ArrayList();
    Object[] obj = get3String(G).toArray();
    for (int i=0;i<obj.length;i++) {
      graph.Vertex[] Varr = (graph.Vertex[]) obj[i];
      graph.Vertex W1 = Varr[2];
      util.Eiffel.assert(W1.size()==5);
      graph.Face F = W1.getAny();
      for (int j=0;j<5;j++) {
        graph.Vertex W = W1.next(F,j).next(W1,1);
        if (W!= Varr[1] && W!= Varr[0] && isPureTriangle(G,W,5)) {
          A.add(new graph.Vertex[]{Varr[0],Varr[1],Varr[2],W});
        }
      }
    }
    return A;
  }

  /**
   * returns an array of all the faces at any of the vertices in Varr
   */
  public static Face[] getFaceAround(graph.Vertex[] Varr) {
    HashSet S = new HashSet();
    for (int i=0;i<Varr.length;i++) {
      graph.Vertex V = Varr[i];
      graph.Face F = V.getAny();
      for (int j=0;j<V.size();j++) {
        graph.Face F2 = V.next(F,j);
        if (!S.contains(F2))
          S.add(F2);
      }
    }
    return (Face[])S.toArray(new Face[0]);
  }

}