package linprog;
import java.util.*;
import java.math.BigDecimal;
import util.*;
import ilog.cplex.example.CpxProblem;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * immutable
 * Contains information about an abstract inequality (not tied to a particular graph)
 * and the domain on which the inequality is valid.
 *
 * Typically these will be parsed in and built with ParseIneq and TemplateBuilder.
 */

public class InequalityTemplate {


  /**
   * Constructor makes a shallow copy of data.  This means that we could
   * modify the internal data by changing the HashMaps that initialize it.
   * Please beware.
   */
  public InequalityTemplate(String name,HashMap VarMin,HashMap VarMax,
    ArrayList coeff, char sense) {
    this.name = name;
    this.VarMin = VarMin;
    this.VarMax = VarMax;
    this.coeff = coeff;
    this.sense = sense;
  }

  private String name;
  private HashMap VarMin = new HashMap(); // VarName->Interval
  private HashMap VarMax = new HashMap(); // VarName->Interval
  private ArrayList coeff = new ArrayList(); // {{VarName,Interval},...}
  private char sense;  // Greater, Less, Equal.

  /**
   * The size is the number of dependent variables.
   * For example, for dihedral angle inequalities on a simplex, the size is 6.
   */
  public int size() {
    util.Eiffel.assert(VarMin.size()==VarMax.size());
    return VarMin.size();
  }

  public String toString() {
    StringBuffer buff = new StringBuffer();
    buff.append("# varnames: \n [");
    for (Iterator I = VarMin.keySet().iterator();I.hasNext();/*--*/) {
      buff.append(I.next().toString());
      buff.append( I.hasNext() ? "," : "]\n" );
    }
    buff.append("# varmin: \n [");
    for (Iterator I = VarMin.keySet().iterator();I.hasNext();/*--*/) {
      buff.append(VarMin.get(I.next()).toString());
      buff.append( I.hasNext() ? "," : "]\n" );
    }
    buff.append("# varmax: \n [");
    for (Iterator I = VarMin.keySet().iterator();I.hasNext();/*--*/) {
      buff.append(VarMax.get(I.next()).toString());
      buff.append( I.hasNext() ? "," : "]\n" );
    }
    for (int i=0;i<coeff.size();i++) {
      Object[] obj = (Object[])coeff.get(i);
      util.Eiffel.assert(obj!=null && obj.length==2);
      buff.append(" ");
      Interval A = (Interval)obj[1];
      if (A.getCenter().signum()>0)
        buff.append("+");
      buff.append(obj[1].toString());
      buff.append(" ");
      buff.append(obj[0].toString());
    }
    char c = (sense==CpxProblem.GreaterEqual ? '>' :
             (sense==CpxProblem.LessEqual ? '<' : '=')
             );
    buff.append(c + " 0\n\n");
    return buff.toString();
  }

  public String getName() {
    return name;
  }


  /**
   * convert the template to an actual inequality.
   * variableDict String varName -> linprog.Variable
   * equations are all homogeneous, so no threshold is required.
   * It is an error if variableDict does not contain all the names of the InequalityTemplate.
   */
  public Inequality toInequality(HashMap variableDict) {
    Inequality ineq = new Inequality(this.toString());
    ineq.setSense(sense);
    ineq.setRHS(new Interval("0"));
    for (int i=0;i<coeff.size();i++) {
      Object[] obj = (Object[]) coeff.get(i);
      util.Eiffel.assert(obj.length==2);
      Object name = obj[0];
      Interval c = (Interval)obj[1];
      Variable V = (Variable)variableDict.get(name);
      util.Eiffel.assert(V!=null,"Cant find "+name+" in "+ineq.toString());
      ineq.addCoeff(V,c);
    }
    return ineq;
  }

  /**
   * A rather messy technical method that gets used extensively.
   * Add dihedral angle inequalities from the factory.
   * Object[] is a list of dictionary data of the length 4*k, k=number of variables,
   *  { String varName, Variable V, Interval minValue, Interval maxValue,....., }
   * Insert null for minValue and maxValue to indicate that V is not a dependent variable.
   * A template matches the data if:
   *  the initial substring of the template's name is 'matchString',
   *  the size of the template is size, and
   *  the method fitsDomain returns true.
   * In this case an inequality is generated from the mapData and template,
   *  and the inequality is returned.
   * Returns null if no match is found.
   */
   public Inequality instantiateTemplate(
    Object[] mapData,String matchString,int size) {
      return instantiateTemplate(mapData,matchString,size,false);
   }

   public Inequality instantiateTemplate(
    Object[] mapData,String matchString,int size,boolean returnNonNull) {
      if (returnNonNull)
        util.Eiffel.assert(this.name.startsWith(matchString),this.name+" "+matchString);
      if (!this.name.startsWith(matchString))
        return null;
      if (returnNonNull)
        util.Eiffel.assert(this.size()==size,""+this.size()+" "+size);
      if (this.size()!=size) {
        //System.out.println(">>wrong size "+size+" "+this.toString());
        return null;
      }
      util.Eiffel.assert((mapData.length % 4) == 0);
      HashMap min = new HashMap();
      HashMap max = new HashMap();
      HashMap dict = new HashMap();
      for (int i=0;i< (mapData.length/4);i++) {
        Object varName = mapData[4*i];
        Object variable = mapData[4*i+1];
        Object minV = mapData[4*i+2];
        Object maxV = mapData[4*i+3];
        util.Eiffel.assert((varName instanceof String),"varName "+varName);
        util.Eiffel.assert((variable instanceof Variable),"variable "+variable+"\n"+varName+"\n"+this.toString());
        util.Eiffel.assert(minV==null || (minV instanceof Interval),"minV "+minV);
        util.Eiffel.assert(maxV==null || (maxV instanceof Interval),"maxV "+maxV);
        if (minV!=null) min.put(varName,minV);
        if (maxV!=null) max.put(varName,maxV);
        dict.put(varName,variable);
     }
     if (returnNonNull) { // DEBUG...
      if (!fitsDomain(min,max)) {
        for (int i=0;i<mapData.length;i++) {
          if (mapData[i]!=null)
            System.out.print(" "+mapData[i].toString());
          else
            System.out.print(" --null-- ");
        }
        System.out.println(this.toString());
      }
      util.Eiffel.assert(fitsDomain(min,max,true));
     } // End DEBUG
    if (!this.fitsDomain(min,max)) {
        //System.out.println(">>doesnt fit domain "+this.toString());
        return null;
    }
    Inequality ineq = this.toInequality(dict);
    if (returnNonNull) {
        util.Eiffel.assert(ineq!=null);
    }
    return ineq;
   }

  /**
   * min : String varName -> Interval (minimum value of variable)
   * max : String varName -> Interval (maximum value of variable)
   * returns true if [min,max] subset [this.VarMin,this.VarMax].
   */
  public boolean fitsDomain(HashMap min,HashMap max) {
    return fitsDomain(min,max,false);
  }
  private boolean fitsDomain(HashMap min,HashMap max,boolean returnTrue) {
    Iterator keys = min.keySet().iterator();
    while (keys.hasNext()) {
      Object key = keys.next();
      BigDecimal B = ((Interval) min.get(key)).min();
      Interval I_thisB = (Interval) VarMin.get(key);
      util.Eiffel.assert(I_thisB!=null, key.toString());
      BigDecimal thisB = I_thisB.max();
      // if thisB > B then false;
      if (returnTrue)
        util.Eiffel.assert(!(thisB.subtract(B).signum()>0),""+thisB+" "+B+" "+key.toString());
      if (thisB.subtract(B).signum()>0)
        return false;
    }
    keys = max.keySet().iterator();
    while (keys.hasNext()) {
      Object key = keys.next();
      BigDecimal B = ((Interval) max.get(key)).max();
      Interval I_thisB = (Interval) VarMax.get(key);
      util.Eiffel.assert(I_thisB!=null);
      BigDecimal thisB = I_thisB.min();
      // if thisB < B then false;
      if (returnTrue)
        util.Eiffel.assert(!(B.subtract(thisB).signum()>0),""+thisB+" "+B+" "+key.toString());
      if (B.subtract(thisB).signum()>0)
        return false;
    }
    return true;
  }

  public static class TestCode extends UnitTest {
    public void test_getMax() {
      Interval A = new Interval("3.0","7");
      InequalityTemplate it = new InequalityTemplate(null,null,null,null,' ');
      assert(A.max().subtract(new BigDecimal("10.0")).signum()==0);
      assert(A.min().subtract(new BigDecimal("-4.0")).signum()==0);
    }
    public void test_fitsDomain() {
      HashMap min = new HashMap();
      min.put("y1",new Interval("3.0","7.0"));
      HashMap max = new HashMap();
      max.put("y1",new Interval("30.0","6.0"));
      ArrayList coeff = new ArrayList();
      coeff.add(new Object[]{"y1",new Interval("3.0")});
      InequalityTemplate it = new InequalityTemplate("dih",min,max,coeff,CpxProblem.GreaterEqual);
      assert(!it.fitsDomain(min,max));
      min = new HashMap();
      max = new HashMap();
      min.put("y1",new Interval("10.0"));
      max.put("y1",new Interval("24.0"));
      assert(it.fitsDomain(min,max));
      min.put("y1",new Interval("9.0"));
      assert(!it.fitsDomain(min,max));
      HashMap map = new HashMap();
    }

  }


}