
/**
 * Title:        Interval.java<p>
 * Description:  An interval arithmetic package based on the java.math.BigDecimal
 *    class.<p>
 * Copyright:    Copyright (c) Thomas C. Hales<p>
 * @author Thomas C. Hales
 * @version 1.0
 */

package linprog;
import java.math.BigDecimal;
import util.*;

/**
 * The Interval Class implements interval arithmetic.  This will be incredibly
 * slow, but it should be entirely reliable.  This should not be used for any
 * high-performance computing.
 * <p>
 * Each interval is represented internally
 * by a pair of BigDecimal objects: radius and center.  The center is the
 * midpoint of the interval and the radius is the half-width, so that the
 * entire interval is [center-radius,center+radius].
 * <p>
 * Arithmetic operations are based on the corresponding operations in the BigDecimal
 * Class.
 * <p>
 * The default precision for all arithmetic operations is 16 decimal places.  All arithmetic
 * operations are rounded (with correct rounding) to 16 decimal places.  This value
 * has been fixed to enforce the immutability of all operations, although it would be
 * easy to modify the code to allow variable precision.
 */

public class Interval {

    private BigDecimal radius;
    private BigDecimal center;
    private static final int scale = 16;  // This sets the rounding precision.

    private static final BigDecimal two = new BigDecimal("2.");
    private static final BigDecimal one = new BigDecimal("1.");

    /**
     * Constructor for positive width intervals.
     * BigDecimals are passed by reference so changes in them change the Interval.
     * A negative radius is permitted, and the constructor takes its absolute value.
     * @param center BigDecimal value of the midpoint
     * @param radius BigDecimal value of half the width of the interval
     */
    public Interval(BigDecimal center, BigDecimal radius) {
        this.center = center;
        this.radius = radius.abs();
    }

    public Interval(BigDecimal center) {
      this.center = center;
      this.radius = I_zero;
    }
    private static final BigDecimal I_zero = new BigDecimal(0);

    /**
     * Constructor for positive width intervals.
     * This constructor is an abbreviation for
     * Interval(BigDecimal(center),BigDecimal(radius));
     * Input of a negative radius is permitted, and the constructor takes its
     * absolute value.
     * @param center String value of the midpoint of the interval
     * @param radius String value of half the width of the interval
     */
    public Interval(String center, String radius) {
        this.center = new BigDecimal(center);
        this.radius = new BigDecimal(radius).abs();
    }

     public String toString() {
      return center.toString()+"`"+radius.toString();
     }


    /**
     * Constructor for 0 width intervals.
     * This constructor is an abbreviation for
     * Interval(BigDecimal(digits),BigDecimal("0"));
     * @param digits String value of the decimal number to be represented
     */
    public Interval(String digits) {
        center = new BigDecimal(digits);
        radius = new BigDecimal("0");
    }


    /**
     * The radius of the interval (exact).  The BigDecimal is immutable, so
     * the Interval Class remains immutable even after the release of the radius.
     */
    public BigDecimal getRadius() {
        return radius;
    }


    /**
     * The center of the interval (exact).  A BigDecimal is immutable, so
     * the Interval Class remains immutable even after the release of the center.
     */
    public BigDecimal getCenter() {
        return center;
    }

    /**
     * The negative of the given interval.
     */
    public Interval negate() {
        return new Interval(center.negate(), radius.abs());
    }

    public BigDecimal max() {
      return center.add(radius);
    }

    public BigDecimal min() {
      return center.subtract(radius);
    }

    /**
     * A.divide(B) is the (normalized) quotient of A by B.
     * @param A Interval value of the divisor
     * @throws ArithmeticException on a divide by an interval containing 0.
     */
    public Interval divide(Interval A) {
        BigDecimal upperLimit = A.center.add(A.radius);
        BigDecimal lowerLimit = A.center.subtract(A.radius);
        if((upperLimit.signum() >= 0) && (lowerLimit.signum() <= 0)) {
            throw new ArithmeticException("divide by zero");
        }
        BigDecimal recipUpperLimit = one.divide(lowerLimit,scale, BigDecimal.ROUND_UP);
        BigDecimal recipLowerLimit = one.divide(upperLimit,scale, BigDecimal.ROUND_DOWN);
        Interval recip = bounds(recipLowerLimit,recipUpperLimit);
        return this.multiply(bounds(recipLowerLimit, recipUpperLimit));
    }

    /**
     * B.add(A) is the sum of B and A.
     * @param A is the Interval value to be added to this Interval
     */
    public Interval add(Interval A) {
        BigDecimal center0 = center.add(A.center);
        BigDecimal radius0 = radius.add(A.radius);
        return new Interval(center0,radius0);
        //BigDecimal upperLimit = center.add(radius).add(A.center).add(A.radius);
        //BigDecimal lowerLimit = center.add(A.center).subtract(A.radius).subtract(radius);
        //return bounds(lowerLimit, upperLimit);
    }

    /**
     * B.multiply(A) is the (normalized) product of B and A.
     * @param A is the Interval multiplier
     */
    public Interval multiply(Interval A) {
        BigDecimal upper = center.add(radius);
        BigDecimal lower = center.subtract(radius);
        BigDecimal Aupper = A.center.add(A.radius);
        BigDecimal Alower = A.center.subtract(A.radius);
        BigDecimal x1 = upper.multiply(Aupper);
        BigDecimal x2 = upper.multiply(Alower);
        BigDecimal x3 = lower.multiply(Aupper);
        BigDecimal x4 = lower.multiply(Alower);
        BigDecimal Bupper = x1.max(x2).max(x3).max(x4);
        BigDecimal Blower = x1.min(x2).min(x3).min(x4);
        return bounds(Blower,Bupper);
    }

    public Interval multiply(BigDecimal A) {
      BigDecimal newCenter = center.multiply(A);
      BigDecimal newRadius = radius.multiply(A).abs();
      return new Interval(newCenter,newRadius);
    }


    /**
     * B.subtract(A) is the interval difference B-A.
     * @param A is the Interval value to be subtracted from this Interval
     */
    public Interval subtract(Interval A) {
        return this.add(A.negate());
    }

    /**
     * Convert the interval into fixed precision (default 16 digit) format.  This process
     * occurs in two steps: (1) convert the center to something close, (2) adjust
     * the radius.
     */
    public Interval normalize() {
        BigDecimal newCenter = center.setScale(scale, BigDecimal.ROUND_DOWN); //any rounding OK
        BigDecimal diff = center.subtract(newCenter).abs();
        BigDecimal newRadius = radius.add(diff).setScale(scale, BigDecimal.ROUND_UP);
        return new Interval(newCenter, newRadius);
    }

    /**
     * Create a normalized Interval object from [lowerLimit,upperLimit].
     * @param lowerLimit BigDecimal value of the lower endpoint of the Interval
     * @param upperLimit BigDecimal value of the upper endpoint of the Interval
     * precondition: upperLimit >= lowerLimit.
     * If the precondition is violated, the behavior of the construct is not defined.
     */
    public static Interval bounds(BigDecimal lowerLimit, BigDecimal upperLimit) {
        if (upperLimit.subtract(lowerLimit).signum()<0)
          throw new ArithmeticException("inverted interval");
        // division by 2 increases the scale by one.
        int bscale = 1+ Math.max(lowerLimit.scale(),upperLimit.scale());
        BigDecimal newCenter = upperLimit.add(lowerLimit).divide(two,bscale, BigDecimal.ROUND_UNNECESSARY);
        BigDecimal newRadius = upperLimit.subtract(lowerLimit).divide(two,bscale, BigDecimal.ROUND_UNNECESSARY);
        return new Interval(newCenter, newRadius).normalize();
    }

  /**
   * Warning: According to BigDecimal conventions
   *  ! new BigDecimal("0.0").equals(new BigDecimal("0.00")),
   * because they have different scales.
   * <p>
   * This is not the convention that the Interval Class follows.
   * Two intervals are equal if the underlying center and radius are the
   * same, viewed as real numbers.  According to these conventions
   *   new Interval("1.0","2.0").equals(new Interval("1.00","2.00"));
   * @param A Interval value to be compared to this Interval
   */
  public boolean equals(Interval A) {
    return (radius.subtract(A.getRadius()).signum()==0 &&
      center.subtract(A.getCenter()).signum()==0);
  }

  /**
   * This is the test code for Interval
   */
  public static class TestCode extends UnitTest {

    static {
      System.out.println(".");
    }

    private boolean bigEquals(BigDecimal A,BigDecimal B) {
      return A.subtract(B).signum()==0;
    }

    public void testEquals() {
      Interval A = new Interval("1.0","3.0");
      Interval B = new Interval("1.0","3.0");
      assert(A.equals(B));
      Interval C = new Interval("3.0","1.0");
      assert(!A.equals(C));
      Interval D = new Interval("1.0","2.0");
      assert(!A.equals(D));
      Interval E = new Interval("0.0","3.0");
      assert(!A.equals(E));
      Interval F = new Interval("0.00","3.00");
      assert(E.equals(F));
    }

    public void testAddition() {
      Interval A = new Interval("1.0","9.0"); // [-8,10]
      Interval B = new Interval("3.0","4.0"); // [-1,7]
      Interval C = new Interval("4.0","13.0"); // [-9,17]
      assert(A.add(B).equals(C));
    }

    public void testNegation() {
      Interval A = new Interval("1.0","3.0"); // [-2,4]
      Interval B = new Interval("-1.0","3.0"); // [-4,2]
      assert(A.negate().equals(B));
    }

    public void testConstructDigits() {
      Interval A = new Interval("1.0","0.0");
      assert(A.equals(new Interval("1.0")));
    }

    public void testConstructorDecimal() {
      Interval A = new Interval("1.0","2.0");
      Interval B = new Interval(new BigDecimal("1.0"),new BigDecimal("2.0"));
      assert(A.equals(B));
    }

    public void testBounds() {
      Interval A = new Interval("3.0","7.0"); // [-4,10].
      Interval B = bounds(new BigDecimal("-4.0"),new BigDecimal("10.0"));
      assert(A.equals(B));
    }

    public void testDivide() {
      Interval A = new Interval("3.0","9.0"); // [-6,12]
      Interval B = new Interval("8.0","4.0"); // [4,12]
      Interval C = bounds(new BigDecimal("-1.5"),new BigDecimal("3"));
      assert(A.divide(B).equals(C),A.divide(B).center+" "+A.divide(B).radius);
      Interval D = new Interval("3.0"); // [3,3];
      Interval ONE = new Interval("1.0");
      Interval near0 = ONE.divide(D).multiply(D).subtract(ONE);
      assert(near0.center.add(near0.radius).signum()>0);
      assert(near0.center.subtract(near0.radius).signum()<0);
      assert(near0.radius.signum()>0);
      try {
        B.divide(A);
        assert(false,"Exception should have been thrown");
        }
      catch (Exception e) {}
    }

    public void testCenter() {
      Interval A = new Interval("3.0","9.0");
      assert(A.getCenter().equals(new BigDecimal("3.0")));
    }

    public void testRadius() {
      Interval A = new Interval("3.0","9.0");
      assert(A.getRadius().equals(new BigDecimal("9.0")));
    }

    public void testMultiply() {
      Interval A = new Interval("1.0","3.0"); // [-2,4]
      Interval B = new Interval("7.0","4.0"); // [3,11]
      Interval C = bounds(new BigDecimal("-22.0"),new BigDecimal("44.0"));
      assert(A.multiply(B).equals(C),A.multiply(B).radius+" "+A.multiply(B).center);
      Interval D = new Interval("-2.0","0.0");
      assert(A.multiply(D).equals(bounds(new BigDecimal("-8.0"),new BigDecimal("4.0"))),"A*D");
      assert(B.multiply(D).equals(bounds(new BigDecimal("-22"),new BigDecimal("-6"))),"A*B");
    }

    public void testMultiplyBigDecimal() {
      Interval A = new Interval("1.0","3.0"); // [-2,4]
      Interval B = new Interval("7.0","4.0"); // [3,11]
      //Interval C = bounds(new BigDecimal("-22.0"),new BigDecimal("44.0"));
      BigDecimal B1 = new BigDecimal(-2.0);
      assert(A.multiply(B1).equals(bounds(new BigDecimal("-8.0"), new BigDecimal("4.0"))),"A*B1");
      assert(B.multiply(B1).equals(bounds(new BigDecimal("-22.0"),new BigDecimal("-6"))),"B*B1");
    }

    public void testSubtract() {
      Interval A = new Interval("1.0","3.0"); // [-2,4]
      Interval B = new Interval("2.0","7.0"); // [-5,9]
      Interval C = new Interval("-1.0","10.0");
      assert(A.subtract(B).equals(C));
    }

    public void testNormalize() {
      // Tests #1 and #2: approximate center, exact radius.
      // test #1: Is the new center machine epsilon from the old center?
      Interval A = new Interval(new BigDecimal(0.1),new BigDecimal("1.0")); // [-0.9+eps,1.1+eps];
      Interval Anorm = A.normalize();
      BigDecimal c = new BigDecimal("0.1");
      assert(bigEquals(Anorm.center,c),"center="+Anorm.center);
      // test #2: Are endpoints within machine epsilon rounded correctly?
      BigDecimal epsilon = new BigDecimal(0.1).subtract(c);
      BigDecimal upperGap = A.center.add(A.radius).subtract(new BigDecimal("1.1"));
      BigDecimal lowerGap = A.center.subtract(A.radius).add(new BigDecimal("0.9"));
      assert(bigEquals(upperGap,lowerGap), "upperGap="+upperGap);
      // test #3: Is the new radius machine epsilon from the old radius?
      Interval Bnorm = new Interval(new BigDecimal("1.0"),new BigDecimal(0.1)).normalize();
      assert(bigEquals(Bnorm.center,new BigDecimal("1.0")),"center="+Bnorm.center);
      assert(bigEquals(
              Bnorm.radius,
              new BigDecimal("0.1").add(epsilon).setScale(Bnorm.scale,BigDecimal.ROUND_UP)
            ),
        "radius="+Bnorm.radius
      );
    }
  }

}
