
/**
 * Title:        Graph00<p>
 * Description:  <p>
 * Copyright:    Copyright (c) Thomas C. Hales<p>
 * Company:      University of Michigan<p>
 * @author Thomas C. Hales
 * @version 1.0
 */
package graph;
import java.util.*;

/**
 * A couple is a pair (v,F) of a vertex and a Face at that vertex.
 * Some algorithms in this package require an Enumeration of all of the
 * couples in a graph.
 */
public class Couple {
    private Vertex V;
    private Face F;

    public Couple(Vertex V, Face F) {
        this.V = V;
        this.F = F;
    }

    public Vertex getV() {
        return V;
    }

    public Face getF() {
        return F;
    }

    /**
     * Helper class for getCouples. check is true when the face "A" contains "V"
     */
    private static class incidence implements Condition {
        private Vertex V;

        public incidence(Vertex V) {
            this.V = V;
        }

        public boolean check(Object A) {
            Face F = (Face)A;
            return (V.next(F, 0) == F); //note: V.next(F,0)==F iff V lies on F.
        }
    }

    /**
     * This returns an enumeration of all Couples in Graph G in a canonical order
     * starting with firstCouple.  The order is canonical in the sense that it
     * only depends on the oriented isomorphism
     * class of G (and firstCouple).  That is, if there is an oriented iso
     *    G <-> G'  sending firstCouple <-> firstCouple', then that bijection will
     * send getCouples <-> getCouples'.
     * <p>
     * This enumeration is useful in
     * finding explicit bijections of isomorphic graphs.
     */

    public static Couple[] getCouples(Couple firstCouple, Graph G) {
        /**
         * Description of algorithm:
         * Return all the couples around V0=firstCouple.V, then all couples around V1, ....
         * The order for couples around firstCouple.V is counterclockwise starting with firstCouple.F.
         * clockwise around V1,...
         * So we need to order the vertices V0,V1,... and pick the first face to use at each vertex
         * and the rest is determined.
         * Order on V0,... = first encountered.  Order on faces at vertex = first encountered.
         */
        distinctFIFO vQueue = new distinctFIFO();
        ConditionGetter fQueue = new ConditionGetter();
        fQueue.add(firstCouple.getF());
        vQueue.put(firstCouple.getV());
        Collection coups = new ArrayList();
        int coupsIndex = 0;
        Vertex V;
        while((V = (Vertex)vQueue.remove()) != null) {
            Face F = (Face)fQueue.get(new incidence(V));
            for(int i = 0;i < V.size();i++) {
                Face Fx = V.next(F, i);
                coups.add(new Couple(V, Fx));
                vQueue.put(Fx.next(V, 1));
                fQueue.add(Fx);
            }
        }
        return (Couple[])coups.toArray(new Couple[coups.size()]);
    }



    /**
     * Test Code for Couple
     */
    public static class Test extends util.UnitTest {

        public void testCouple() {
            String S = Formatter.testString; //see formatterString.gif
            Graph G = Graph.getInstance(new Formatter(S));
            Vertex V = null;
            int coupleCount = 0;
            for(Enumeration E = G.vertexEnumeration();E.hasMoreElements(); /*--*/) {
                Vertex W = (Vertex)E.nextElement();
                coupleCount += W.size();
                if(W.size() == 3)
                    V = W;
            }
            assert(V.size() == 3);
            Face F = V.getAny();
            while(F.size() != 5)
                F = V.next(F, 1);
            assert(F.size() == 5);
            Couple C = new Couple(V, F);
            Couple[] list = Couple.getCouples(C, G);
            //compute the expected number of return values;
            assert(list.length == coupleCount);
            assert(list[0].getV() == C.getV());
            assert(list[0].getF() == C.getF());
            //check integrity of each couple.
            for(int i = 0;i < list.length;i++) {
                V = list[i].getV();
                F = list[i].getF();
                assert(V.next(F, 0) == F);
                assert(F.next(V, 0) == V);
            }
            //check that all couples are distinct.
            Hashtable table = new Hashtable(); // { V-> set of F }
            for(Enumeration E = G.vertexEnumeration();E.hasMoreElements(); /*--*/) {
                V = (Vertex)E.nextElement();
                table.put(V, new HashSet());
            }
            for(int i = 0;i < list.length;i++) {
                V = list[i].getV();
                F = list[i].getF();
                HashSet H = (HashSet)table.get(V);
                assert(!H.contains(F));
                H.add(F);
            }
        }
    }
}
