#include "cplex_CpxProblem.h"
#include <cplex.h>
#include <stdlib.h>

/*  Overview:
 *
 *
 *  This file contains the implementation of the native methods of the CPLEX
 *  Java class CpxProblem.  Every native method has at least two parameters:
 *  JNIEnv *jenv, and jobject jobj.  jenv is a pointer to the Java virtual
 *  machine.  It provides the interface functions needed for accessing data
 *  within java objects.  jobj is a reference to the java object whose method
 *  is being called, in this case the CpxProblem object.  Access to individual
 *  members of jobj are done through functions defined for JNIEnv.  We will
 *  briefly discuss these functions when we use them the first time in the C
 *  functions that follow.
 *
 *  Before starting this discussion let us highlight some general principles
 *  that are important to consider when extending existing or adding new stubs.
 *
 *  1. This implementation is based on JNI as defined for JDK 1.2.1.  See
 *     http://java.sun.com/products/jdk/1.2/docs/guide/jni/index.html.
 *
 *  2. Since Java is a platform independent language it defines the sizes of
 *     basic data types to be the same across all platforms.  In contrast the C
 *     types for pointers are different between platforms.  Therefore all data
 *     received from a Java object must be cast from the Java type to the
 *     corresponding C type and vice versa.  For an example see the function
 *     Java_ilog_cplex_example_CpxProblem_getX() which uses an implicit cast of
 *     C type double to jdouble.
 *
 *  3. The location of java objects is dynamic, as the Java garbage collector
 *     is allowed to move them at any time.  Therefore, generally one must go
 *     through the functional API mentioned above to access data in Java
 *     objects.  This would yield bad performance when accessing array elements
 *     one by one.  To overcome this problem, JNI offers the possibility of
 *     "pinning" an array to a location.  This would return a C pointer to
 *     the data in the Java array and instruct the java garbage collector to
 *     no longer move this array.  The pointer returned by pinning can be
 *     used like any C array until the pointer is released again.  It is
 *     important to release the pointer after using it for the Java virtual
 *     machine to run smoothly.  Be sure not to use the pointer after releasing
 *     it, because this would result in sporadically appearing errors that are
 *     very hard to track down.  The Java function names for pinning an array
 *     depends on the array type;  for an array of type X the name is
 *     GetXArrayElements().  Similarily the Java function name for releasing an
 *     array of type X is ReleaseXArrayElements().
 */ 


/*
 *  The CpxProblem object keeps the CPLEX environment and lp pointer in members
 *  of the Java type long; see CpxProblem.java.  The Java type long has been
 *  selected in order to assure that pointer will fit on both, 32 and 64 bit
 *  architectures.  To access them from the stubs, *  function getEnvLp() is
 *  used.  The procedure is the following:
 *
 *  1. get a reference xcplex to the Java class CpxProblem by calling
 *     FindClass().
 *  2. get an accessor to the "lp" and "env" member of CpxProblem objects by
 *     calling GetFieldID().
 *  3. get the value of the members lp and env from object obj by calling
 *     GetLongField().
 *  4. use a cast to convert the jlong values to the CPLEX environment and lp
 *     pointers.
 */
static void getEnvLP (JNIEnv *jenv, jobject obj,
                      CPXENVptr *env_p, CPXLPptr *lp_p)
{
   jclass    xcplex = (*jenv)->FindClass(jenv, "ilog/cplex/example/CpxProblem");
   jfieldID  xlp    = (*jenv)->GetFieldID(jenv, xcplex,  "lp", "J");
   jfieldID  xenv   = (*jenv)->GetFieldID(jenv, xcplex, "env", "J");
   jlong     env    = (*jenv)->GetLongField(jenv, obj, xenv);
   jlong     lp     = (*jenv)->GetLongField(jenv, obj, xlp);

   *env_p = (CPXENVptr)(long)env;
   *lp_p  = (CPXLPptr)(long)lp;
}


/* Whenever a CPLEX call returns a nonzero error code a Java exception of type
 * CpxException as defined in CpxException.java is generated.  To do this an
 * object of type CpxException must be created from within the C stubs.  This
 * is done with the following steps:
 *
 * 1. Get a reference exclass to the CpxException class in the java virtual
 *    machine by calling FindClass().
 * 2. Get an ID to the constructor method for CpxException objects by calling
 *    GetMethodID().
 * 3. Create a Java string from the CPLEX error message by calling
 *    NewStringUTF().
 * 4. Create the exception object by calling NewObject().
 * 5. Throw the exception object by calling Throw().
 */
static void throwStatus (JNIEnv *jenv, CPXENVptr env, int status)
{
   if ( status ) {
      jclass    exclass = (*jenv)->FindClass(jenv,
                                             "ilog/cplex/example/CpxException");
      jmethodID ctor    = (*jenv)->GetMethodID(jenv, exclass,
                                               "<init>",
                                               "(ILjava/lang/String;)V");

      char    msg[1024];
      jstring jstr;
      jobject jex;

      CPXgeterrorstring (env, status, msg);
      jstr = (*jenv)->NewStringUTF(jenv, msg);
      jex  = (*jenv)->NewObject(jenv, exclass, ctor, (jint)status, jstr);

      (*jenv)->Throw(jenv, jex);
   }
}

static int getSparseVector(JNIEnv *jenv, jobject sv,
                           jint **ind_p, jdouble **val_p)
{
   jclass   jsv  = (*jenv)->FindClass(jenv,
                                      "ilog/cplex/example/CpxSparseVector");
   jfieldID jnum = (*jenv)->GetFieldID(jenv, jsv, "num", "I");

   if ( ind_p != NULL && val_p != NULL ) {
      jfieldID jind = (*jenv)->GetFieldID(jenv, jsv, "ind", "[I");
      jfieldID jval = (*jenv)->GetFieldID(jenv, jsv, "val", "[D");
      jobject  xval = (*jenv)->GetObjectField(jenv, sv, jval);
      jobject  xind = (*jenv)->GetObjectField(jenv, sv, jind);

      *ind_p = (*jenv)->GetIntArrayElements(jenv, xind, NULL);
      *val_p = (*jenv)->GetDoubleArrayElements(jenv, xval, NULL);
   }

   return ((*jenv)->GetIntField(jenv, sv, jnum));
}


static void releaseSparseVector(JNIEnv *jenv, jobject sv,
                               jint *ind, jdouble *val)
{
   jclass   jsv  = (*jenv)->FindClass(jenv,
                                      "ilog/cplex/example/CpxSparseVector");
   jfieldID jind = (*jenv)->GetFieldID(jenv, jsv, "ind", "[I");
   jfieldID jval = (*jenv)->GetFieldID(jenv, jsv, "val", "[D");
   jobject  xval = (*jenv)->GetObjectField(jenv, sv, jval);
   jobject  xind = (*jenv)->GetObjectField(jenv, sv, jind);

   (*jenv)->ReleaseIntArrayElements(jenv, xind, ind, 0);
   (*jenv)->ReleaseDoubleArrayElements(jenv, xval, val, 0);
}



/*
 * Class:     CpxProblem
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_init (JNIEnv *jenv,
                                                                jobject obj)
{
   int       status = 0;
   jlong     j_lp, j_env;

   CPXLPptr  lp   = 0;
   CPXENVptr env  = 0;

   jclass    xcplex = (*jenv)->FindClass(jenv, "ilog/cplex/example/CpxProblem");
   jfieldID  xlp    = (*jenv)->GetFieldID(jenv, xcplex, "lp",  "J");
   jfieldID  xenv   = (*jenv)->GetFieldID(jenv, xcplex, "env", "J");

   env = CPXopenCPLEXdevelop(&status);
   if ( status )  goto TERMINATE;

   lp = CPXcreateprob (env, &status, "java.lp");
   if ( status )  goto TERMINATE;

   j_lp  = (jlong)(long)lp;
   j_env = (jlong)(long)env;

   (*jenv)->SetLongField(jenv, obj,  xlp, j_lp);
   (*jenv)->SetLongField(jenv, obj, xenv, j_env);

TERMINATE:

   throwStatus(jenv, env, status);

}

/*
 * Class:     CpxProblem
 * Method:    finit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_finit (JNIEnv *jenv,
                                                                 jobject obj)
{
   int       status = 0;
   CPXLPptr  lp   = 0;
   CPXENVptr env  = 0;

   getEnvLP (jenv, obj, &env, &lp);

   status = CPXfreeprob (env, &lp);
   if ( status )  goto TERMINATE;

   status = CPXcloseCPLEX (&env);
   if ( status )  goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    getItcnt
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_ilog_cplex_example_CpxProblem_getItcnt (
   JNIEnv *jenv,
   jobject obj)
{
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);

   return (CPXgetitcnt(env, lp));
}

/*
 * Class:     CpxProblem
 * Method:    getObj
 * Signature: ()D
 */
JNIEXPORT jdouble JNICALL Java_ilog_cplex_example_CpxProblem_getObj (
   JNIEnv *jenv,
   jobject obj)
{
   int         status = 0;
   double      objval;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);

   if ( CPXgetprobtype (env, lp) == CPXPROB_MIP ) {
      status = CPXgetmipobjval (env, lp, &objval);
   }
   else {
      status = CPXgetobjval (env, lp, &objval);
   }
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);

   return (objval);
}

/*
 * Class:     CpxProblem
 * Method:    write
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_write (JNIEnv *jenv,
                                                                 jobject obj,
                                                                 jstring str)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;
   const char* file = (*jenv)->GetStringUTFChars(jenv, str, 0);

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXwriteprob (env, lp, (char*)file, NULL);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}


/*
 * Class:     CpxProblem
 * Method:    read
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_read (JNIEnv *jenv,
                                                                jobject obj,
                                                                jstring str)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;
   const char* file = (*jenv)->GetStringUTFChars(jenv, str, 0);

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXreadcopyprob (env, lp, (char*)file, NULL);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    dualopt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_dualopt (
   JNIEnv *jenv,
   jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXdualopt (env, lp);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    primopt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_primopt (
   JNIEnv *jenv,
   jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXprimopt (env, lp);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    baropt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_baropt (JNIEnv *jenv,
                                                                  jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXbaropt (env, lp);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    hybbaropt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_hybbaropt (
   JNIEnv *jenv,
   jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXhybbaropt (env, lp, CPXALG_PRIMAL);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}


/*
 * Class:     CpxProblem
 * Method:    hybnetopt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_hybnetopt (
   JNIEnv *jenv,
   jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXhybnetopt (env, lp, CPXALG_DUAL);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}


/*
 * Class:     CpxProblem
 * Method:    mipopt
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_mipopt (JNIEnv *jenv,
                                                                  jobject obj)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, obj, &env, &lp);
   status = CPXmipopt (env, lp);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}


/*
 * Class:     CpxProblem
 * Method:    getStatus
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_ilog_cplex_example_CpxProblem_getStatus (
   JNIEnv* jenv,
   jobject jobj)
{
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   return (CPXgetstat(env, lp));
}


/*
 * Class:     CpxProblem
 * Method:    getDblQuality
 * Signature: (I)D
 * Note Cplex Version 6.5 or higher
 */
/*
JNIEXPORT jdouble JNICALL Java_ilog_cplex_example_CpxProblem_getDblQuality (
   JNIEnv* jenv,
   jobject jobj,
   jint    which)
{
   int         status = 0;
   double      val;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   status = CPXgetdblquality(env, lp, &val, which);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);

   return (val);
}
*/

/*
 * Class:     CpxProblem
 * Method:    getIntQuality
 * Signature: (I)I
 * Note: Cplex Version 6.5 or higher (tch)
 */
/*
JNIEXPORT jint JNICALL Java_ilog_cplex_example_CpxProblem_getIntQuality (
   JNIEnv* jenv,
   jobject jobj,
   jint    which)
{
   int         status = 0;
   int         val;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   status = CPXgetintquality(env, lp, &val, which);
   if ( status ) goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);

   return (val);
}
*/


/*
 * Class:     CpxProblem
 * Method:    getNCols
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_ilog_cplex_example_CpxProblem_getNCols (
   JNIEnv* jenv,
   jobject jobj)
{
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   return (CPXgetnumcols(env, lp));
}


/*
 * Class:     CpxProblem
 * Method:    getNRows
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_ilog_cplex_example_CpxProblem_getNRows (
   JNIEnv* jenv,
   jobject jobj)
{
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   return (CPXgetnumrows(env, lp));
}


/*
 * Class:     CpxProblem
 * Method:    getBase
 * Signature: ([I[I)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_getBase (
   JNIEnv*   jenv,
   jobject   jobj,
   jintArray xcstat,
   jintArray xrstat)
{
   int       status = 0;
   int       i;
   jint*     jrstat;
   jint*     jcstat;
   int       ncols, nrows;
   int       *cstat = NULL;
   int       *rstat = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   ncols = CPXgetnumcols(env, lp);
   nrows = CPXgetnumrows(env, lp);

   if ( (*jenv)->GetArrayLength(jenv, xcstat) < ncols ||
        (*jenv)->GetArrayLength(jenv, xrstat) < nrows   ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   cstat = (int*) malloc (ncols * sizeof(int));
   rstat = (int*) malloc (nrows * sizeof(int));

   if (cstat == NULL || rstat == NULL) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }

   status = CPXgetbase(env, lp, cstat, rstat);
   if ( status ) goto TERMINATE;

   jrstat = (*jenv)->GetIntArrayElements(jenv, xrstat, NULL);
   for (i = 0; i < nrows; ++i)  jrstat[i] = rstat[i];
   (*jenv)->ReleaseIntArrayElements(jenv, xrstat, jrstat, 0);

   jcstat = (*jenv)->GetIntArrayElements(jenv, xcstat, NULL);
   for (i = 0; i < ncols; ++i)  jcstat[i] = cstat[i];
   (*jenv)->ReleaseIntArrayElements(jenv, xcstat, jcstat, 0);

TERMINATE:

   throwStatus(jenv, env, status);

   free (cstat);
   free (rstat);
}


/*
 * Class:     CpxProblem
 * Method:    getX
 * Signature: ([D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_getX (
   JNIEnv*     jenv,
   jobject     jobj,
   jdoubleArray xx)
{
   int       status = 0;
   int       i;
   jdouble*  jx;
   int       ncols;
   double*   x   = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   ncols = CPXgetnumcols(env, lp);

   if ( (*jenv)->GetArrayLength(jenv, xx) < ncols ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   x = (double*) malloc (ncols * sizeof(double));

   if ( x == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   if ( CPXgetprobtype (env, lp) == CPXPROB_MIP ) {
      status = CPXgetmipx(env, lp, x, 0, ncols-1);
   }
   else {
      status = CPXgetx(env, lp, x, 0, ncols-1);
   }
   if ( status ) goto TERMINATE;

   jx = (*jenv)->GetDoubleArrayElements(jenv, xx, NULL);
   for (i = 0; i < ncols; ++i)  jx[i] = x[i];
   (*jenv)->ReleaseDoubleArrayElements(jenv, xx, jx, 0);

TERMINATE:

   throwStatus(jenv, env, status);

   free (x);
}


/*
 * Class:     CpxProblem
 * Method:    getDj
 * Signature: ([D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_getDj (
   JNIEnv*      jenv,
   jobject      jobj,
   jdoubleArray xdj)
{
   int       status = 0;
   int       i;
   jdouble*  jdj;
   int       ncols;
   double*   dj  = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   ncols = CPXgetnumcols(env, lp);

   if ( (*jenv)->GetArrayLength(jenv, xdj) < ncols ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   dj = (double*) malloc (ncols * sizeof(double));

   if ( dj == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   status = CPXgetdj(env, lp, dj, 0, ncols-1);
   if ( status ) goto TERMINATE;

   jdj = (*jenv)->GetDoubleArrayElements(jenv, xdj, NULL);
   for (i = 0; i < ncols; ++i)  jdj[i] = dj[i];
   (*jenv)->ReleaseDoubleArrayElements(jenv, xdj, jdj, 0);

TERMINATE:

   throwStatus(jenv, env, status);

   free (dj);
}

/*
 * Class:     CpxProblem
 * Method:    getPi
 * Signature: ([D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_getPi (
   JNIEnv*      jenv,
   jobject      jobj,
   jdoubleArray xpi)
{
   int       status = 0;
   int       i;
   jdouble*  jpi;
   int       nrows;
   double*   pi  = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   nrows = CPXgetnumrows(env, lp);

   if ( (*jenv)->GetArrayLength(jenv, xpi) < nrows ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   pi = (double*) malloc (nrows * sizeof(double));

   if ( pi == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   status = CPXgetpi(env, lp, pi, 0, nrows-1);
   if ( status ) goto TERMINATE;

   jpi = (*jenv)->GetDoubleArrayElements(jenv, xpi, NULL);
   for (i = 0; i < nrows; ++i)  jpi[i] = pi[i];
   (*jenv)->ReleaseDoubleArrayElements(jenv, xpi, jpi, 0);

TERMINATE:

   throwStatus(jenv, env, status);

   free (pi);
}

/*
 * Class:     CpxProblem
 * Method:    getSlack
 * Signature: ([D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_getSlack (
   JNIEnv*      jenv,
   jobject      jobj,
   jdoubleArray xslack)
{
   int       status = 0;
   int       i;
   jdouble*  jslack;
   int       nrows;
   double*   slack  = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   nrows = CPXgetnumrows(env, lp);

   if ( (*jenv)->GetArrayLength(jenv, xslack) < nrows ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   slack = (double*) malloc (nrows * sizeof(double));

   if ( slack == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   if ( CPXgetprobtype (env, lp) == CPXPROB_MIP ) {
      status = CPXgetmipslack(env, lp, slack, 0, nrows-1);
   }
   else {
      status = CPXgetslack(env, lp, slack, 0, nrows-1);
   }
   if ( status ) goto TERMINATE;

   jslack = (*jenv)->GetDoubleArrayElements(jenv, xslack, NULL);
   for (i = 0; i < nrows; ++i)  jslack[i] = slack[i];
   (*jenv)->ReleaseDoubleArrayElements(jenv, xslack, jslack, 0);

TERMINATE:

   throwStatus(jenv, env, status);

   free (slack);
}


/*
 * Class:     CpxProblem
 * Method:    setObjSense
 * Signature: (C)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_setObjSense (
   JNIEnv* jenv,
   jobject jobj,
   jint    xsense)
{
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   CPXchgobjsen(env, lp, xsense);
}

/*
 * Class:     CpxProblem
 * Method:    setDoubleParam
 * Signature: (ID)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_setDoubleParam (
   JNIEnv* jenv,
   jobject jobj,
   jint    which,
   jdouble val)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   status = CPXsetdblparam(env, which, val);
   if ( status )  goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}

/*
 * Class:     CpxProblem
 * Method:    setIntParam
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_setIntParam (
   JNIEnv* jenv,
   jobject jobj,
   jint    which,
   jint    val)
{
   int         status = 0;
   CPXLPptr    lp   = 0;
   CPXENVptr   env  = 0;

   getEnvLP (jenv, jobj, &env, &lp);
   status = CPXsetintparam(env, which, val);
   if ( status )  goto TERMINATE;

TERMINATE:

   throwStatus(jenv, env, status);
}



/*
 * Class:     CpxProblem
 * Method:    addCols
 * Signature: ([D[LCpxSparseVector;[D[D[C)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_addCols (
   JNIEnv*      jenv,
   jobject      jlp,
   jdoubleArray xobj,
   jobjectArray xcols,
   jdoubleArray xlb,
   jdoubleArray xub,
   jcharArray   xtype)
{
   int       status = 0;
   int       num, i, n;
   jdouble*  jobj;
   jdouble*  jlb;
   jdouble*  jub;
   jchar*    jtype;
   double*   obj  = NULL;
   double*   lb   = NULL;
   double*   ub   = NULL;
   char*     type = NULL;
   int*      ind  = NULL;
   int*      beg  = NULL;
   double*   val  = NULL;
   CPXLPptr  lp   = 0;
   CPXENVptr env  = 0;
   int       isMIP = 0;

   getEnvLP (jenv, jlp, &env, &lp);

   num = 0;
   n   = (*jenv)->GetArrayLength(jenv, xcols);
   if ( n ) {
      jobject sv;
      for (i = 0; i < n; ++i) {
         sv = (*jenv)->GetObjectArrayElement(jenv, xcols, i);
         if ( sv ) num += getSparseVector(jenv, sv, 0, 0);
      }
   }

   n = (*jenv)->GetArrayLength(jenv, xobj);

   if ( (*jenv)->GetArrayLength(jenv, xub)   != n ||
        (*jenv)->GetArrayLength(jenv, xlb)   != n ||
        (*jenv)->GetArrayLength(jenv, xtype) != n   ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   obj  = (double*) malloc (n * sizeof(double));
   lb   = (double*) malloc (n * sizeof(double));
   ub   = (double*) malloc (n * sizeof(double));
   type = (char*) malloc (n * sizeof(char));

   if ( obj  == NULL ||
        lb   == NULL ||
        ub   == NULL ||
        type == NULL   ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   jtype = (*jenv)->GetCharArrayElements(jenv, xtype, NULL);
   jobj  = (*jenv)->GetDoubleArrayElements(jenv, xobj, NULL);
   jlb   = (*jenv)->GetDoubleArrayElements(jenv, xlb,  NULL);
   jub   = (*jenv)->GetDoubleArrayElements(jenv, xub,  NULL);
   for (i = 0; i < n; ++i) {
      obj[i]  = jobj[i];
      lb[i]   = jlb[i];
      ub[i]   = jub[i];
      type[i] = jtype[i];
      if ( type[i] != 'C' ) isMIP = 1;
   }
   (*jenv)->ReleaseCharArrayElements(jenv, xtype, jtype, 0);
   (*jenv)->ReleaseDoubleArrayElements(jenv, xobj, jobj, 0);
   (*jenv)->ReleaseDoubleArrayElements(jenv, xlb,  jlb,  0);
   (*jenv)->ReleaseDoubleArrayElements(jenv, xub,  jub,  0);

   if ( num == 0 ) {
      if ( isMIP ) status = CPXnewcols (env, lp, n, obj, lb, ub, type, NULL);
      else         status = CPXnewcols (env, lp, n, obj, lb, ub, NULL, NULL);
   }
   else {
      jint*    jind;
      jdouble* jval;
      int k, l, m;
      jobject sv;

      beg  = (int   *) malloc (n   * sizeof(int));
      ind  = (int   *) malloc (num * sizeof(int));
      val  = (double*) malloc (num * sizeof(double));
      if ( beg == NULL || ind == NULL || val == NULL ) {
         status = CPXERR_NO_MEMORY;
         goto TERMINATE;
      }

      for (i = k = 0; i < n; ++i) {
         beg[i] = k;
         sv = (*jenv)->GetObjectArrayElement(jenv, xcols, i);
         if ( sv ) {
            l  = getSparseVector(jenv, sv, &jind, &jval);
            for (m = 0; m < l; ++m, ++k) {
               ind[k] = jind[m];
               val[k] = jval[m];
            }
            releaseSparseVector(jenv, sv, jind, jval);
         }
      }
      status  = CPXaddcols (env, lp, n, num, obj, beg, ind, val, lb, ub, NULL);
   }

TERMINATE:

   throwStatus(jenv, env, status);

   free (ub);
   free (lb);
   free (obj);
   free (type);
   free (beg);
   free (ind);
   free (val);
}

/*
 * Class:     CpxProblem
 * Method:    addRows
 * Signature: ([LCpxSparseVector;[I[D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_addRows (
   JNIEnv*      jenv,
   jobject      jlp,
   jobjectArray xrows,
   jcharArray   xsense,
   jdoubleArray xrhs)
{
   int       status = 0;
   int       num, i, n;
   jdouble*  jrhs;
   jchar*    jsense;
   double*   rhs   = NULL;
   char*     sense = NULL;
   int*      ind   = NULL;
   int*      beg   = NULL;
   double*   val   = NULL;
   CPXLPptr  lp    = 0;
   CPXENVptr env   = 0;

   getEnvLP (jenv, jlp, &env, &lp);

   num = 0;
   n   = (*jenv)->GetArrayLength(jenv, xrows);
   if ( n ) {
      jobject sv;
      for (i = 0; i < n; ++i) {
         sv = (*jenv)->GetObjectArrayElement(jenv, xrows, i);
         if ( sv ) num += getSparseVector(jenv, sv, 0, 0);
      }
   }

   n = (*jenv)->GetArrayLength(jenv, xrhs);

   if ( (*jenv)->GetArrayLength(jenv, xsense) != n ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   rhs   = (double*) malloc (n * sizeof(double));
   sense = (char  *) malloc (n * sizeof(char));

   if ( rhs == NULL || sense == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   jsense = (*jenv)->GetCharArrayElements(jenv, xsense, NULL);
   jrhs   = (*jenv)->GetDoubleArrayElements(jenv, xrhs, NULL);
   for (i = 0; i < n; ++i) {
      sense[i] = jsense[i];
      rhs[i]   = jrhs[i];
   }
   (*jenv)->ReleaseCharArrayElements(jenv, xsense, jsense, 0);
   (*jenv)->ReleaseDoubleArrayElements(jenv, xrhs, jrhs, 0);

   if ( num == 0 ) {
      status = CPXnewrows (env, lp, n, rhs, sense, NULL, NULL);
   }
   else {
      jint*    jind;
      jdouble* jval;
      int k, l, m;
      jobject sv;

      beg  = (int   *) malloc (n   * sizeof(int));
      ind  = (int   *) malloc (num * sizeof(int));
      val  = (double*) malloc (num * sizeof(double));
      if ( beg == NULL || ind == NULL || val == NULL ) {
         status = CPXERR_NO_MEMORY;
         goto TERMINATE;
      }

      for (i = k = 0; i < n; ++i) {
         beg[i] = k;
         sv = (*jenv)->GetObjectArrayElement(jenv, xrows, i);
         if ( sv ) {
            l  = getSparseVector(jenv, sv, &jind, &jval);
            for (m = 0; m < l; ++m, ++k) {
               ind[k] = jind[m];
               val[k] = jval[m];
            }
            releaseSparseVector(jenv, sv, jind, jval);
         }
      }
      status  = CPXaddrows (env, lp, 0, n, num, rhs, sense, beg, ind, val,
                            NULL, NULL);
   }

TERMINATE:

   throwStatus(jenv, env, status);

   free (rhs);
   free (sense);
   free (beg);
   free (ind);
   free (val);
}


/*
 * Class:     CpxProblem
 * Method:    setCoefs
 * Signature: ([I[I[D)V
 */
JNIEXPORT void JNICALL Java_ilog_cplex_example_CpxProblem_setCoefs (
   JNIEnv*      jenv,
   jobject      jobj,
   jintArray    xrind,
   jintArray    xcind,
   jdoubleArray xval)
{
   int       status = 0;
   int       i, num;
   jint*     jrind;
   jint*     jcind;
   jdouble*  jval;
   int       *cind = NULL;
   int       *rind = NULL;
   double    *val = NULL;
   CPXLPptr  lp  = 0;
   CPXENVptr env = 0;

   getEnvLP (jenv, jobj, &env, &lp);

   num = (*jenv)->GetArrayLength(jenv, xval);
   if ( (*jenv)->GetArrayLength(jenv, xcind) != num ||
        (*jenv)->GetArrayLength(jenv, xrind) != num   ) {
      status = CPXERR_ARRAY_TOO_LONG;
      goto TERMINATE;
   }
   
   cind = (int*)    malloc (num * sizeof(int));
   rind = (int*)    malloc (num * sizeof(int));
   val  = (double*) malloc (num * sizeof(double));

   if ( cind == NULL || rind == NULL || val == NULL ) {
      status = CPXERR_NO_MEMORY;
      goto TERMINATE;
   }

   jrind = (*jenv)->GetIntArrayElements(jenv, xrind, NULL);
   for (i = 0; i < num; ++i)  rind[i] = jrind[i];
   (*jenv)->ReleaseIntArrayElements(jenv, xrind, jrind, 0);

   jcind = (*jenv)->GetIntArrayElements(jenv, xcind, NULL);
   for (i = 0; i < num; ++i)  cind[i] = jcind[i];
   (*jenv)->ReleaseIntArrayElements(jenv, xcind, jcind, 0);

   jval = (*jenv)->GetDoubleArrayElements(jenv, xval, NULL);
   for (i = 0; i < num; ++i)  val[i] = jval[i];
   (*jenv)->ReleaseDoubleArrayElements(jenv, xval, jval, 0);

   status = CPXchgcoeflist (env, lp, num, rind, cind, val);

TERMINATE:

   throwStatus(jenv, env, status);

   free (cind);
   free (rind);
   free (val);
}
