package ilog.cplex.example;

/** Cplex solver and problem class.
 *  Each object of this class will allocate its own CPXLP and CPXENV object.
 *  As a consequence each object grabs its own license which is freed only
 *  upon destruction of the CpxProblem Java object.  Since this may occur
 *  any time when the Java garbage collector decides to, it is advised to call
 *  method CpxProblem.dispose() explicitly.
 *
 *  The user is assumed to be familiar with the C callable library.
 */
public class CpxProblem {
  private long env;
  private long lp;

  static {
    try {
       System.loadLibrary("cplexjava60");
     }
     catch (UnsatisfiedLinkError link) {
       System.out.println(link.toString());
       System.out.println("try setting LD_LIBRARY_PATH to the current directory");
     }
  }

  private native void  init() throws CpxException;
  private native void  finit() throws CpxException;

  /** constant for inifinity bounds. */
  public final static double Infinity      = 1.0E+20;

  /** constant for minimzation objective sense */
  public final static int    Minimize      =  1;

  /** constant for maximiztion objective sense */
  public final static int    Maximize      = -1;

  /** constant for equality constraint sense */
  public final static char   Equal         = 'E';

  /** constant for less or equal constraint sense */
  public final static char   LessEqual     = 'L';

  /** constant for greater or equal constraint sense */
  public final static char   GreaterEqual  = 'G';

  /** constant for integer variable type */
  public final static char   Integer       = 'I';

  /** constant for continuous variable type */
  public final static char   Continuous    = 'C';

  /** Creates one CPLEX solver and problem representation object.
    * The CPLEX object are created as C object that are linked to Java through
    * JNI (the Java Native Interface).  Hence, CpxProblem objects can only be
    * created on machines that have libjavacplex65.so or javacplex65.dll
    * installed on UNIX or Windows plattforms, respectively.  A valid CPLEX
    * license must also exist on the machine.
    */
  public               CpxProblem() throws CpxException {
    super(); init();
  }

  protected void       finalize() throws java.lang.Throwable {
     super.finalize(); finit();
  }

  /** Destructor.  It is advised to call this function explicitly to destroy a
   * CpxProblem object, rather than relying on the Java garbage collector to
   * call the destructor at some arbirary time in the future.  Calling it
   * explicitly will ensure that the CPLEX license is released at this time.
   */
  public void          dispose() throws java.lang.Throwable {
     finalize();
  }


  /** Read an LP, MPS or SAV file from the file system.
   * @param file        the name of the file to be read.
   */
  public native void   read  (String file) throws CpxException;

  /** Write a LP, MPS or SAV format file to the file system.
   * @param file        the name of the file to be written.  The extension of
   * the file name determines what format the file will be written in.
   */
  public native void   write (String file) throws CpxException;

  /** Call the dual simplex algorithm to optimize the problem in the CpxProblem
   * object.  The loaded problem must not contain any integer variables.
   */
  public native void   dualopt() throws CpxException;

  /** Call the primal simplex algorithm to optimize the problem in the CpxProblem
   * object.  The loaded problem must not contain any integer variables.
   */
  public native void   primopt() throws CpxException;

  /** Call the barrier algorithm to optimize the problem in the CpxProblem
   * object.  The loaded problem must not contain any integer variables.
   */
  public native void   baropt() throws CpxException;

  /** Call the barrier algorithm followed by primal crossover to optimize the
   * problem in the CpxProblem * object.  The loaded problem must not contain
   * any integer variables.  */
  public native void   hybbaropt() throws CpxException;

  /** Call the network simplex algorithm followed by dual simplex to optimize
   * the problem in the CpxProblem object.  The loaded problem must not contain
   * any integer variables.  */
  public native void   hybnetopt() throws CpxException;

  /** Call the MIP optimizer to solve the problem in the CpxProblem object.
   * The loaded problem must contain at least one integer variable.  */
  public native void   mipopt() throws CpxException;


  /** return the objective value of the solution currently stored in the
   * CpxProblem object. */
  public native double getObj() throws CpxException;

  /** return the number of iteration from the last simplex run. */
  public native int    getItcnt() throws CpxException;

  /** return the current solution status. */
  public native int    getStatus() throws CpxException;

  /** return the number of columns of the problem stored in the CpxProblem
   * object. */
  public native int    getNCols() throws CpxException;

  /** return the number of rows of the problem stored in the CpxProblem
   * object. */
  public native int    getNRows() throws CpxException;


  /** return the objective sense of the problem stored in the CpxProblem
   * object. */
  public native void   setObjSense(int sense) throws CpxException;


  /** Add rows to the problem stored in the CpxProblem object.
   * @param rows        an array of sparse vectors containing the column
   *                    indices and associated nonzero coefficients of the rows
   *                    to be added.
   * @param senses      an array of characters CpxProblem.Equal,
   *                    CpxProblem.LessEqual, or CpxProblem.GreaterEqual that
   *                    indicate the constraint sense of the rows to be added.
   * @param rhs         an array of right hand side values for the rows to be
   *                    added.
   */
  public native void   addRows(CpxSparseVector[] rows,
                               char[]            senses,
                               double[]          rhs) throws CpxException;

  /** Add columns to the problem stored in the CpxProblem object.
   * @param obj         an array of objective value coefficients for the
   *                    columns to be added.
   * @param cols        an array of sparse vectors containing the row indices
   *                    and associated nonzero coefficients of the columns
   *                    to be added.
   * @param lb          an array of lower bound values for the columns to be
   *                    added.
   * @param ub          an array of upper bound values for the columns to be
   *                    added.
   * @param type        an array of variable type characters CpxProblem.Integer
   *                    or CpxProblem.Continous for the columns to be added.
   */
  public native void   addCols(double[]          obj,
			       CpxSparseVector[] cols,
			       double[]          lb,
			       double[]          ub,
			       char[]            type) throws CpxException;

  /** Set nonzero coefficients in the problem stored in the CpxProblem object.
   * @param rind        an array of row indices for the coefficients to be set.
   * @param cind        an array of column indices for the coefficients to be set.
   * @param val         an array of values for the coefficients to be set.
   */
  public native void   setCoefs(int[] rind, int[] cind, double[] val)
    throws CpxException;


  /** Obtain basis information of the current solution.
   * @param cstat       an array where the column status values are copied to.
   * @param rstat       an array where the row status values are copied to.
   */
  public native void   getBase (int[] cstat, int[] rstat)
    throws CpxException;

  /** Obtain the primal solution vector of the current solution.
   * @param x           an array where the primal solution values are copied to.
   */
  public native void   getX    (double[] x) throws CpxException;

  /** Obtain the dual solution vector of the current solution.
   * @param x           an array where the dual solution values are copied to.
   */
  public native void   getPi   (double[] x) throws CpxException;

  /** Obtain the reduced cost vector of the current solution.
   * @param x           an array where the reduced cost values are copied to.
   */
  public native void   getDj   (double[] x) throws CpxException;

  /** Obtain the vector of slack values for the current solution.
   * @param x           an array where the slack values are copied to.
   */
  public native void   getSlack(double[] x) throws CpxException;


  /** Obtain double-valued quality information about the current solution.
   * @param which       an integer indicating the type of information to be
   *                    returned.
   */
  public native double getDblQuality(int which) throws CpxException;

  /** Obtain integer-valued quality information about the current solution.
   * @param which       an integer indicating the type of information to be
   *                    returned.
   */
  public native int    getIntQuality(int which) throws CpxException;


  /** Set a CPLEX integer-valued paramter.
   * @param which       an integer indicating the integer parameter to be set.
   * @param value       the new parameter value.
   */
  public native void   setIntParam(int which, int value) throws CpxException;


  /** Set a CPLEX double-valued paramter.
   * @param which       an integer indicating the double parameter to be set.
   * @param value       the new parameter value.
   */
  public native void   setDoubleParam(int which, double value) throws CpxException;
}
