package ilog.cplex.example;

/** Class for representing sparse vectors.
 *
 */
public class CpxSparseVector {
   int      num;
   int[]    ind;
   double[] val;

   /** create a sparse vector with no nonzeros.  The sparse vector is ready to
    * hold 8 nonzeros without requiring a reallocation of internal buffers.
    */
   public       CpxSparseVector() {
                   num = 0;
                   ind = new int[8];
                   val = new double[8];
                }

   /** create a sparse vector with no nonzeros.  The sparse vector is ready to
    * hold max nonzeros without requiring a reallocation of internal buffers.
    * @param max        number of nonzeros that can be added without
    *                   reallocation of internal buffers.
    */
   public       CpxSparseVector(int max) {
                   num = 0;
                   ind = new int[max];
                   val = new double[max];
                }

   /** return number of nonzeros currently in the sparse vector */
   public int   getNNZ() { return num; }

   /** return index of the i'th nonzero */
   public int   getIndex(int i) {
                   if ( i < 0 || i >= num ) {
                      throw new IndexOutOfBoundsException();
                   }
                   return ind[i];
                }

   /** return coefficient value of the i'th nonzero */
   public double
                getValue(int i) {
                   if ( i < 0 || i >= num ) {
                      throw new IndexOutOfBoundsException();
                   }
                   return val[i];
                }

   /** add a nonzero with index i and coefficient value v */
   public void  add(int i, double v) {
                   if ( num >= ind.length ) {
                      int[]    xind = new int[2 * ind.length];
                      double[] xval = new double[2 * ind.length];
		      System.arraycopy(ind, 0, xind, 0, num);
		      System.arraycopy(val, 0, xval, 0, num);
                      ind = xind;
                      val = xval;
                   }
                   ind[num] = i;
                   val[num] = v;
                   ++num;
                }

   /** add a nonzero with index i and coefficient value v */
   public void  add(double v, int i) { add(i,v); }

   /** remove n'th nonzero from sparse vector */
   public void  remove(int n) {
                   if ( n < 0 || n >= num ) {
                      throw new IndexOutOfBoundsException();
                   }
                   --num;
                   for (int i = n; i < num; ++i) {
                      ind[i] = ind[i+1];
                      val[i] = val[i+1];
                   }
                }

   /** return number of nonzero with index j in the sparse vector
     * @return  number of nonzero with index j in the sparse vector.  If no such
     *          nonzero is available in the sparse vector -1 will be returned.
     */
   public int   find(int j) {
                   int i = num-1;
                   while (i >= 0  && ind[i] != j) --i;
                   return i;
                }
};
