package linprog;
import graph.Face;
import java.util.*;
import ilog.cplex.example.CpxProblem;
import ilog.cplex.example.CpxException;
import java.math.BigDecimal;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * Routines to generate a linear program.  The hope is to put everything that
 * is needed in this class that needs to be visible outside the package.
 * This isn't literally true, there is a method that returns the MetricGraph,
 *  if any modifications need to be made to it directly.
 *
 * This is shallow immutable.
 */

public class GraphLP {

  // links to outside classes.
  private final MetricGraphVariables MGV;
  private final LPdata data = new LPdata();
  public Enumeration variableEnumeration() {
    return data.variableEnumeration();
  }

  public HashMap getSolutionVariable() throws CpxException {
    return data.getSolutionVariable();
  }

  public double getSolution(Variable V) throws CpxException {
    return ((Double) data.getSolutionVariable().get(V)).doubleValue();
  }

  // static data.
  private static final Parameters P = ParameterFactory.get();
  //private static final InequalityFactory ineqFac = new InequalityFactory(P);
  private static final InequalityTemplate[] templateList;
  static {
    ArrayList template = new ArrayList();
    for (int i=0;i<P.ineqFiles().length;i++) {
      String text="";
      try {
        text = IO.getString(P.ineqPath+P.ineqFiles()[i]);
      }
      catch (java.io.IOException e) {
        e.printStackTrace();
      }
      ParseIneq parse = new ParseIneq(text);
      Node N = parse.parseInequalityList(parse.Tokenize());
      TemplateBuilder b = new TemplateBuilder(N);
      template.addAll(b.readInequalityList());
    }
    templateList = (InequalityTemplate[])template.toArray(new InequalityTemplate[0]);
  }


  public GraphLP(MetricGraph MG) {
    this.MGV = new MetricGraphVariables(MG);
    initLPdataInstance();
    addStandardIneq();
  }


  /**
   * returns underlying metric graph.
   * This returns a shared copy, so any changes become changes to the
   * MetricGraph underlying the GraphLP object.
   * For instance, constraints can be added to the linear optimization by
   *  modifying the properties of the MetricGraph.
   */
  public MetricGraph getMetricGraph() {
    return MGV.getMetricGraph();
  }

  public MetricGraphVariables getMetricGraphVariables() {
    return MGV;
  }

  public Inequality getSigmaAsInequality() {
    return InequalityFactory.sigmaOver8pt(MGV.getSigma());
  }

  /**
   * @parameter sense = CpxProblem.Minimize or CpxProblem.Maximize
   */
  public BigDecimal runObjectiveV(graph.Vertex V, int sense, BigDecimal thresholdDefault) throws CpxException {
    Objective obj = ObjectiveFactory.varLimit(MGV.varV(V), sense, thresholdDefault);
    Inequality ineq = this.getSigmaAsInequality();
    data.addInequality(ineq);
    data.setObjective(obj);
    data.verifyThreshold(Parameters.threshold);
    BigDecimal answer = data.optimum();
    data.removeInequality(ineq);
    return answer;
  }

  public BigDecimal runObjectiveE(graph.Vertex V1, graph.Vertex V2, int sense) throws CpxException {
    Objective obj = ObjectiveFactory.varLimit(MGV.varE(V1,V2), sense, null);
    Inequality ineq = this.getSigmaAsInequality();
    data.addInequality(ineq);
    data.verifyThreshold(Parameters.threshold);
    data.setObjective(obj);
    BigDecimal answer = data.optimum();
    data.removeInequality(ineq);
    return answer;
  }

  public BigDecimal runObjectiveSigma() throws CpxException {
    Objective obj = ObjectiveFactory.sigma(MGV.getSigma());
    data.setObjective(obj);
    data.verifyThreshold(Parameters.threshold);
    BigDecimal answer = data.optimum();
    return answer;
  }

  public BigDecimal runObjectivePerim(graph.Face F, int sense,BigDecimal defaultThreshold) throws CpxException {
    ArrayList list = new ArrayList();
    graph.Vertex V = F.getAny();
    for (int i=0;i<F.size();i++) {
      list.add(MGV.varE(F.next(V,i),F.next(V,i+1)));
    }
    Objective obj = ObjectiveFactory.varLimit(list.iterator(), sense, defaultThreshold);
    Inequality ineq = this.getSigmaAsInequality();
    data.setObjective(obj);
    data.addInequality(ineq);
    data.verifyThreshold(Parameters.threshold);
    BigDecimal answer = data.optimum();
    data.removeInequality(ineq);
    return answer;
  }

  /**
   * @parameter LPdata, has-side effects, new inequalities are added.
   */
  private void addAngleSumToData() {
      for (Enumeration E = MGV.vertexEnumeration();E.hasMoreElements();/*--*/) {
        graph.Vertex V = (graph.Vertex) E.nextElement();
        Face F = V.getAny();
        ArrayList varList = new ArrayList();
        for (int i=0;i<V.size();i++) {
          Variable t = MGV.varAngle(V.next(F,i),V);
          util.Eiffel.assert(t!=null);
          varList.add( t );
        }
        data.addInequality(InequalityFactory.sum2Pi(varList.iterator()));
      }
  }


  /*
  private void addSmallQRTetToData() {
    for (Enumeration E = MGV.getMetricGraph().faceEnumeration();E.hasMoreElements();) {
      graph.Face F = (graph.Face)E.nextElement();
      if (F.size()!=3)
        continue;
      graph.Vertex V = F.getAny();
      graph.Vertex U = F.next(V,1);
      graph.Vertex W = F.next(V,2);
      if (MGV.getMetricGraph().isBIGQRTET(F)) {
        ArrayList list = new ArrayList();
        list.add(MGV.varE(V,U));
        list.add(MGV.varE(U,W));
        list.add(MGV.varE(V,W));
        Inequality ineq = InequalityFactory.lowerBoundI(list.iterator(),new Interval("6.25","0"));
        data.addInequality(ineq);
      }
      if (this.getMetricGraph().isSmallQRTET(F)) {
        ArrayList list = new ArrayList();
        list.add(MGV.varE(V,U));
        list.add(MGV.varE(U,W));
        list.add(MGV.varE(V,W));
        Inequality ineq = InequalityFactory.upperBoundI(list.iterator(),new Interval("6.25","0"));
        data.addInequality(ineq);
      }
    }
  }
  */

  /**
   * has side-effects on data.
   */
  private void addSolidDefToData() {
    for (Enumeration E = MGV.faceEnumeration();E.hasMoreElements();/*--*/) {
      graph.Face F = (graph.Face) E.nextElement();
      Variable Fvar = MGV.varSolid(F);
      graph.Vertex V = F.getAny();
      ArrayList varList = new ArrayList();
      for (int i=0;i<F.size();i++) {
        Variable t = MGV.varAngle(F,F.next(V,i));
        util.Eiffel.assert(t!=null);
        varList.add(t);
      }
      data.addInequality(InequalityFactory.solid(varList.iterator(),Fvar));
    }
  }

  /**
   * Add the quoin inequalities.
   * This has got 3 dependent variables T(y1,y2,y3).
   */
  private void addQuoin() {
    for (Iterator I = MGV.getQuoinKeys();I.hasNext();/*--*/) {
      //System.out.print("*"); // DEBUG
      Object[] key = (Object[])I.next();
      graph.Vertex V1 = MGV.edge0(key);
      graph.Vertex V2 = MGV.edge1(key);
      for (int i=0;i<templateList.length;i++) {
        InequalityTemplate template = templateList[i];
        if (template.size()!=3) continue;
        if (!template.getName().startsWith("quo")) continue;
        Inequality ineq = template.instantiateTemplate(
            new Object[] {
              "unit",P.unit,null,null,
              "y1",MGV.varV(V1),new Interval(MGV.getMinHeight(V1)),new Interval(MGV.getMaxHeight(V1)),
              "y2",MGV.varV(V2),new Interval(MGV.getMinHeight(V2)),new Interval(MGV.getMaxHeight(V2)),
              "y3",MGV.varE(V1,V2),MGV.getMinEdge(V1,V2),MGV.getMaxEdge(V1,V2),
              "quo",MGV.varQ(V1,V2),null,null
              },
            "quo",3);
        if (ineq!=null)
          data.addInequality(ineq);
      }
    }

  }

  /**
   * Add the SPIII Misc Inequalities, (Group 4)
   */
  private void addMiscIII() {
    MetricGraph MG = MGV.getMetricGraph();
    //1. add SPIII_Lemma53[b]
    Object[] objectList = GraphStructure.getPureTriangleStrings(MG);
    for (int i=0;i<objectList.length;i++) {
        graph.Vertex[] verList = (graph.Vertex[])objectList[i];
        Face[] faceList = GraphStructure.getFaceAround(verList);
        ArrayList A = new ArrayList();
        ArrayList B = new ArrayList();
        for (int j=0;j<faceList.length;j++) {
          A.add(MGV.varSigma(faceList[j]));
          B.add(MGV.varSolid(faceList[j]));
        }
        data.addInequality(InequalityFactory.SPIII_Lemma53(A.iterator(),B.iterator(),verList.length));
        data.addInequality(InequalityFactory.SPIII_Lemma53b(A.iterator(),verList.length));
    }
    for (Enumeration E = MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex) E.nextElement();
      //2. add sum033 on type (4,0).
      if (GraphStructure.isPureTriangle(MG,V,4)) {
        Face[] faceList = GraphStructure.getFaceAround(new graph.Vertex[]{V});
        ArrayList A = new ArrayList();
        for (int i=0;i<faceList.length;i++) {
          A.add(MGV.varSigma(faceList[i]));
        }
        data.addInequality(InequalityFactory.sum033(A.iterator()));
      }
      //3. add sum5 on type {5,0}
      if (GraphStructure.isPureTriangle(MG,V,5)) {
        Face[] faceList = GraphStructure.getFaceAround(new graph.Vertex[]{V});
        ArrayList A = new ArrayList();
        ArrayList B = new ArrayList();
        for (int i=0;i<faceList.length;i++) {
          A.add(MGV.varSigma(faceList[i]));
          B.add(MGV.varSolid(faceList[i]));
        }
        data.addInequality(InequalityFactory.sum5(A.iterator(),B.iterator()));
      }

    }
  }

  private static void addData(ArrayList data,String name,Variable V) {
    data.add(name);
    data.add(V);
    data.add(null);
    data.add(null);
  }

  private Object[] dihHelper(boolean isTriangle,Angle dih,Angle dih2,Angle dih3,Adih adih,Solid sol,Sigma sig,
          graph.Vertex W1,graph.Vertex W2,graph.Vertex W3) {
    ArrayList mapData = new ArrayList();
    addData(mapData,"dih",dih);
    addData(mapData,"dih2",dih2);
    addData(mapData,"dih3",dih3);
    addData(mapData,"adih",adih);
    addData(mapData,"sol",sol);
    addData(mapData,"sigma",sig);
    addData(mapData,"unit",P.unit);
    // Java documentation states that "addAll" adds objects in the proper order.
    mapData.addAll(rowV("y1",W1));
    mapData.addAll(rowV("y2",W2));
    mapData.addAll(rowV("y3",W3));
    if (isTriangle) mapData.addAll(rowE("y4",W2,W3));
    mapData.addAll(rowE("y5",W1,W3));
    mapData.addAll(rowE("y6",W1,W2));
    return mapData.toArray();
    }

  private void addExcept() {
    for (Enumeration E = MGV.faceEnumeration();E.hasMoreElements();/*--*/) {
      Face F = (Face) E.nextElement();
      if (F.size()<5)
        continue;
      util.Eiffel.assert(F.size()<7);
      Interval uppersigma = null;
      //1. sigma < ...
      if (F.size()==5)
        uppersigma = new Interval("-0.05704");
      else
        uppersigma = new Interval("-0.11408");

      data.addInequality(InequalityFactory.upperBound(MGV.varSigma(F),uppersigma));
      //2. tau > .....
      Interval lowertau = null;
      if (F.size()==5)
        lowertau = new Interval("0.27113");
      else
        lowertau = new Interval("0.41056");
      Inequality ineq = new Inequality("tauLB");
      addTauToLHS(MGV.varSigma(F),MGV.varSolid(F),ineq);
      ineq.setSense(CpxProblem.GreaterEqual);
      ineq.setRHS(lowertau);
      ineq.addCoeff(P.threshold,ineq.getRHS());
      data.addInequality(ineq);
    }
  }



  private void excessAtExcept() {
    Interval lowertau5 = new Interval("0.27113");
    Interval lowertau6 = new Interval("0.41056");
    //1. build list of vertices of type {4,0,1}, {3,1,1}, and {3,0,2};
    ArrayList relevantV = new ArrayList();
    for (Enumeration E = MGV.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex)E.nextElement();
      if (V.size()!=5)
        continue;
      if (V.faceCount(3,3)==4 && V.faceCount(5,8)==1) {
        relevantV.add(V);
        continue;
      }
      if (V.faceCount(3,3)==3 && V.faceCount(5,8)>0)
        relevantV.add(V);
    }
    //2. build all subsets of relevantV
    for (int i=0;i<Math.pow(2,relevantV.size());i++) {
      ArrayList subset = new ArrayList();
      boolean subsetOK = true;
      for (int j=0;j<relevantV.size();j++) {
        if (0==((i>>j) % 2)) {
          //System.out.print("0");
          graph.Vertex V = (graph.Vertex) relevantV.get(j);
          subset.add(V);
        }
      }
      //2.a build inequality and save.
      if (!subsetOK)
        continue;
      if (subset.size()==0)
        continue;
      Inequality ineq = new Inequality("excessAt");
      ineq.setSense(CpxProblem.GreaterEqual);
      Interval rhs = P.I_zero;
      Interval pt14 = new Interval("0.0775231"); // a little less than 1.4pt
      Interval pt15 = new Interval("0.08306"); // a little less than 1.5pt.
      Set faceSet = new HashSet();
      for (int j=0;j<subset.size();j++) {
        graph.Vertex V = (graph.Vertex)subset.get(j);
        util.Eiffel.assert(V.size()==5);
        graph.Face F = V.getAny();
        for (int k=0;k<V.size();k++) {
          faceSet.add(V.next(F,k));
        }
        if (V.faceCount(3,3)==3 && V.faceCount(5,8)>0) {
          // add 1.4pt
          rhs = rhs.add(pt14);

        }
        else if (V.faceCount(3,3)==4 && V.faceCount(5,8)==1) {
          // add 1.5pt
          rhs = rhs.add(pt15);
        }
        else
          util.Eiffel.assert(false,"Unreachable state");
      }
      Set vertexSet = new HashSet(subset);
      for (Iterator I = faceSet.iterator();I.hasNext();/*--*/) {
        Face F = (Face)I.next();
        graph.Vertex V = F.getAny();
        for (int j=0;j<F.size();j++) {
          graph.Vertex W = F.next(V,i);
          if (vertexSet.contains(W) && vertexSet.contains(F.next(V,i+1)))
            subsetOK=false;
          if (F.size()==4 && vertexSet.contains(W) && vertexSet.contains(F.next(V,i+2)))
            subsetOK=false;
        }
        Sigma sig = MGV.varSigma(F);
        Solid sol = MGV.varSolid(F);
        addTauToLHS(sig,sol,ineq);
        if (F.size()==5) {
          rhs = rhs.add(lowertau5);
        }
        if (F.size()==6) {
          rhs = rhs.add(lowertau6);
        }
        util.Eiffel.assert(F.size()<7);
      }//Iterator I over faces.
      ineq.setRHS(rhs);
      ineq.addCoeff(P.threshold,ineq.getRHS());
      if (subsetOK) {
        data.addInequality(ineq);
        //System.out.println(ineq.toString());
      }
    }//for i running over subsets.

  } //

  private void addTauToLHS(Sigma sig,Solid sol,Inequality ineq) {
    ineq.addCoeff(sig,P.I_mOne);
    ineq.addCoeff(sol,new Interval("0.100444572"));
  }

  /**
   * Add all the dihedral angle inequalities, as well as adih inequalities.
   * The dihedral angle inequalites are 6D for triangular faces, and 5D otherwise
   *  (there is a y4 variable on triangles, but not on other faces).
   *
   * This has got triple sum (face,vertex,template), so it might be slow.
   */
  private void addDih() {
    for (Enumeration E = MGV.faceEnumeration();E.hasMoreElements();/*--*/) {
      Face F = (Face) E.nextElement();
      boolean isTriangle = (F.size()==3);
      graph.Vertex V = F.getAny();
      Inequality ineq = null;
      Solid sol = MGV.varSolid(F);
      Sigma sig = MGV.varSigma(F);
      //System.out.println("There are "+templateList.length+" ineqs in template list");
      for (int i=0;i<F.size();i++) {
        graph.Vertex W1 = F.next(V,i);
        graph.Vertex W2 = F.next(V,i+1);
        graph.Vertex W3 = F.next(V,i-1);
        Angle dih = MGV.varAngle(F,W1);
        Angle dih2 =MGV.varAngle(F,W2);
        Angle dih3 =MGV.varAngle(F,W3);
        Adih adih = MGV.varAdih(F,W1);
        Interval ymin = new Interval(MGV.getMinHeight(W1));
        Object[] firstData=dihHelper(isTriangle,dih,dih2,dih3,adih,sol,sig,W1,W2,W3);
        if (getMetricGraph().isFATQUAD(F)) {
          graph.Vertex W7 = F.next(V,i+2);
          ArrayList A = new ArrayList();
          for (int j=0;j<firstData.length;j++) {
            A.add(firstData[j]);
          }
          addData(A,"y7",MGV.varV(W7));
          addData(A,"y8",MGV.varE(W7,W3));
          addData(A,"y9",MGV.varE(W7,W2));
          firstData = A.toArray();
        }
        /* We repeat the code here, with the symmetry y2<->y3, y5<->y6.
             This extra code allows us to avoid writing every dihedral inequality twice.
          */
        Object[] secondData=dihHelper(isTriangle,dih,dih3,dih2,adih,sol,sig,W1,W3,W2);
        for (int j=0;j<templateList.length;j++) {
          InequalityTemplate template = templateList[j];
          if (isTriangle) {
            MetricGraph G = this.getMetricGraph();
            if (G.isQRTET(F)) {
              if (i==0) addIneq(template,firstData,"sym3",6);
              addIneq(template,firstData,"dih3",6);
              addIneq(template,secondData,"dih3",6);
              if (G.isSmallQRTET(F)) {
                if (i==0) addIneq(template,firstData,"smallsym3",6);
                addIneq(template,firstData,"smalldih3",6);
                addIneq(template,secondData,"smalldih3",6);
              }
              if (G.isBIGQRTET(F) && i==0)
                addIneq(template,firstData,"bigsym3",6);
            }
            if (G.isFLATQUARTER(W1,F))
              addIneq(template,firstData,"flat",6);
            if (G.isUPRIGHT(W1,F))
              addIneq(template,firstData,"up",6);
          }
          if (isTriangle)
            continue;
          if (F.size()!=4)
            continue;
          // so now it is a quad.
          addIneq(template,firstData,"quad",5);
          MetricGraph G = getMetricGraph();
          if (G.isFATQUAD(F))
            addIneq(template,firstData,"fatquad",5);


        } //j
      } //i
    } //E
  }

  private void addIneq(InequalityTemplate template,Object[] vardata,String start,int numIndepVar) {
    Inequality ineq = template.instantiateTemplate(vardata,start,numIndepVar);
    if (ineq!=null)
      data.addInequality(ineq);
  }

  /**
   * helper for addDih
   */
   private ArrayList rowV(String name,graph.Vertex V) {
    ArrayList list = new ArrayList();
    list.add(name);
    list.add(MGV.varV(V));
    list.add(new Interval(MGV.getMinHeight(V)));
    list.add(new Interval(MGV.getMaxHeight(V)));
    return list;
   }

   /**
    * helper for addDih
    */
   private ArrayList rowE(String name,graph.Vertex V1,graph.Vertex V2) {
    ArrayList list = new ArrayList();
    list.add(name);
    list.add(MGV.varE(V1,V2));
    list.add(MGV.getMinEdge(V1,V2));
    list.add(MGV.getMaxEdge(V1,V2));
    return list;
   }

   /**
    * number of edges of F of length at least eM
    */
  private int kCount(graph.Face F) {
    int k=0;
    graph.Vertex V = F.getAny();
    for (int i=0;i<F.size();i++) {
      graph.Vertex V1 = F.next(V,i);
      graph.Vertex V2 = F.next(V,i+1);
      Interval length = MGV.getMinEdge(V1,V2);
      // length >= eM
      if (length.min().subtract(P.eM().max()).signum()>=0)
        k++;
    }
    return k;
  }

  private void addVorDef() {
    for (Enumeration E=this.getMetricGraph().faceEnumeration();E.hasMoreElements();/*--*/) {
      Face F = (Face)E.nextElement();
      if (F.size()!=4)
        continue;
      if (!MGV.getMetricGraph().isFATQUAD(F))
        continue;
      Sigma sig = MGV.varSigma(F);
      Solid sol = MGV.varSolid(F);
      graph.Vertex V = F.getAny();
      ArrayList AdihList = new ArrayList();
      ArrayList QuoinList = new ArrayList();
      for (int i=0;i<F.size();i++) {
        graph.Vertex W = F.next(V,i);
        AdihList.add(MGV.varAdih(F,W));
        QuoinList.add(MGV.varQ(W,F.next(W,1)));
        QuoinList.add(MGV.varQ(W,F.next(W,-1)));
        util.Eiffel.assert(MGV.varQ(W,F.next(W,1))!=null);
        util.Eiffel.assert(MGV.varQ(W,F.next(W,-1))!=null);
      }
      Iterator AdihOnFace = AdihList.iterator();
      Iterator QuoinOnFace = QuoinList.iterator();
      Inequality ineq = InequalityFactory.vor0(AdihOnFace,QuoinOnFace,sol,sig);
      util.Eiffel.assert(ineq!=null);
      data.addInequality(ineq);
    }
  }

  private void addHeight() {
    MetricGraph MG = MGV.getMetricGraph();
    for (Enumeration E = MGV.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex)E.nextElement();
      data.addInequality(InequalityFactory.lowerBound(MGV.varV(V),new Interval(MG.getMinHeight(V))));
      data.addInequality(InequalityFactory.upperBound(MGV.varV(V),new Interval(MG.getMaxHeight(V))));
      graph.Face F = V.getAny();
      for (int i=0;i<V.size();i++) {
        graph.Vertex W = V.next(F,i).next(V,1);
        data.addInequality(InequalityFactory.lowerBound(MGV.varE(V,W),MG.getMinEdge(V,W)));
        data.addInequality(InequalityFactory.upperBound(MGV.varE(V,W),MG.getMaxEdge(V,W)));
      }
    }
  }

   private void addStandardIneq() {
    addHeight();
    addVorDef();
    addAngleSumToData();
    addSolidDefToData();
    //addSmallQRTetToData();
    addDih(); //dih and adih and sigma and sol ...
    addQuoin();
    addMiscIII();
    addExcept();
    excessAtExcept();
   }



  /**
   * Helper for getLPdataInstance
   * values of M are added as Variables to data.
   * Warning: has a side-effect on data.
   * @precondition values of varI can be cast to (Variable)
   */
  private void addVariables( Iterator varI) {
    while (varI.hasNext()) {
      Variable X = (Variable)varI.next();
      data.addVariable(X,P.threshold);
    }
  }

  /**
   * Generates LPdata with Variable names.
   * Doesn't add inequalities or objective;
   * All new LPdata objects should start here.
   */
  private void initLPdataInstance() {
    data.addVariable(P.threshold,null);
    data.addVariable(P.unit,P.threshold);
    data.addVariable(P.cvolume,P.threshold);
    addVariables(MGV.getV());
    addVariables(MGV.getSigma());
    addVariables(MGV.getEdge());
    addVariables(MGV.getQuoin());
    addVariables(MGV.getAngle());
    addVariables(MGV.getSolid());
    addVariables(MGV.getAdih());
    quarteredQuads();
  }

  public boolean quarteredQuads() {
    MetricGraph MG = MGV.getMetricGraph();
    //1. find one upright diagonal.
    Object[] VF = MG.getUpright();
    if (VF==null)
      return false;
    graph.Vertex V = (graph.Vertex)VF[0];
    graph.Face F = (graph.Face)VF[1];
    util.Eiffel.assert(V.size()==4);
    Sigma sig = new Sigma("U");
    Angle[] A = {new Angle("U1"),new Angle("U2"),new Angle("U3"),new Angle("U4")};
    Solid sol = new Solid("sU");
    ArrayList angles = new ArrayList();
    data.addVariable(sig,P.threshold);
    data.addVariable(sol,P.threshold);
    for (int i=0;i<A.length;i++) {
        data.addVariable(A[i],P.threshold);
        angles.add(A[i]);
    }
    // sum solF = sum A1 - 2Pi
    data.addInequality(InequalityFactory.solid(angles.iterator() ,sol));
    // sum sigmaF = sig
    ArrayList sigmas = new ArrayList();
    for (int i=0;i<V.size();i++) {
        sigmas.add(MGV.varSigma(V.next(F,i)));
    }
    data.addInequality(InequalityFactory.sumIdentity(sigmas.iterator(),sig));
    // sum dih = dihU
    util.Eiffel.assert(A.length==4);
    for (int i=0;i<A.length;i++) {
      ArrayList dihs = new ArrayList();
      graph.Vertex W = V.next(F,i).next(V,1);
      Face G1 = V.next(F,i);
      Face G2 = W.next(G1,-1);
      util.Eiffel.assert(G1.next(V,0)==V && G2.next(V,0)==V);
      util.Eiffel.assert(G2.next(W,0)==W && G2.next(W,0)==W);
      dihs.add(MGV.varAngle(G1,W));
      dihs.add(MGV.varAngle(G2,W));
      data.addInequality(InequalityFactory.sumIdentity(dihs.iterator(),A[i]));
    }
    // add standard quad inequalities.
    for (int i=0;i<4;i++) {
      ArrayList firstData = new ArrayList();
      addData(firstData,"sigma",sig);
      addData(firstData,"sol",sol);
      addData(firstData,"unit",P.unit);
      addData(firstData,"dih",A[i % A.length]);
      addData(firstData,"dih2",A[(i+1) % A.length]);
      for (int j=0;j<templateList.length;j++) {
        InequalityTemplate template = templateList[j];
        addIneq(template,firstData.toArray(),"quad",5);
      }
    }
    return true;
  }

  public String toString() {
    return data.toString();
  }

  public static void main(String[] args) {
    System.out.println("hi");
    for (int i=0;i<templateList.length;i++) {
      System.out.println(templateList[i].toString());
    }
  }


}