package linprog;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 * Company:
 * @author Thomas C. Hales
 * @version 2001.0
 */

import java.math.BigDecimal;
import ilog.cplex.example.CpxProblem;
import util.*;
import java.util.Iterator;

/**
 * represents a Row in a linear program.
 * The inequality is sum ai xi # rhs,
 * where sum ai xi is the coefficient list, # is the sense of the inequality.
 */
public class Inequality extends CoefficientList {

  private char sense;
  private Interval rhs;
  private Object name;

  /**
   * CpxProblem.Equal CpxProblem.LessEqual CpxProblem.GreaterEqual
   */
  public void setSense(char sense) {
    this.sense = sense;
  }

  public String getName() {
    return name.toString();
  }

  /**
   * Add two inequalities.  If A and B have opposite senses, then
   * B is first multiplied by -1 to reverse the sense.
   */
  public static Inequality add(Inequality A, Inequality B) {
    //1. if only one has sense '=', make it the second.
    if (A.getSense() == CpxProblem.Equal) {
      Inequality swap = A;
      A = B; B = swap;
    }
    if (A.getSense()!= B.getSense() && B.getSense()!= CpxProblem.Equal) {
        B = B.scaledIneq(new Interval("-1.0"));
    }
    //2. initialize
    Inequality C =  A.scaledIneq(new Interval("1.0"));
    C.setRHS(A.getRHS().add(B.getRHS()));
    //3. add together.
    Variable V;
    for (Iterator I = B.getVariableIterator(); I.hasNext();/*--*/) {
      V = (Variable) I.next();
      C.addCoeff(V,B.getCoefficient(V));
    }
    return C;
  }

  /**
   * has side-effects. Adds inequality B to "this"
   * @precondition: this.sense==Equal || B.sense==Equal || this.sense==B.sense.
   */
  public void augmentBy(Inequality B) {
    util.Eiffel.precondition(this.sense==CpxProblem.Equal ||
        B.sense==CpxProblem.Equal ||
        B.sense==this.sense);
    if (this.sense==CpxProblem.Equal)
      this.sense = B.sense;
    Variable V;
    for (Iterator I=B.getVariableIterator();I.hasNext();/*--*/) {
      V = (Variable) I.next();
      addCoeff(V,B.getCoefficient(V));
    }
    rhs = rhs.add(B.rhs);
  }

  /**
   * If the inequality is sum ai xi < b, the gap is an upper bound
   * on b - sum ai xi.
   * If the inequality is sum ai xi > b, the gap is a lower bound
   * on b - sum ai xi
   * precondition: sense is LessEqual or GreaterEqual.
   */
  public BigDecimal gap() {
    util.Eiffel.assert(this.getSense()==CpxProblem.LessEqual || this.getSense()==CpxProblem.GreaterEqual);
    Interval A = rhs;
    for (Iterator I = this.getVariableIterator();I.hasNext();/*--*/){
      Variable V = (Variable) I.next();
      Interval coeff = this.getCoefficient(V);
      Interval bound = Interval.bounds(V.getLowerBound(),V.getUpperBound());
      A = A.subtract(coeff.multiply(bound));
    }
    if (this.getSense()==CpxProblem.LessEqual) {
      return A.getCenter().add(A.getRadius());
    }
    return A.getCenter().subtract(A.getRadius());
  }

  /**
   * subtract the coefficients of the CoefficientList from the given inequality.
   * Note that the sense of the inequality is not considered: always subtract.
   */
  public Inequality diminishBy(CoefficientList coeff) {
    Inequality C = this.scaledIneq(new Interval("1.0")); // copy.
    for (Iterator I=coeff.getVariableIterator();I.hasNext();/*--*/) {
      Variable V = (Variable) I.next();
      C.addCoeff(V,coeff.getCoefficient(V).negate());
    }
    return C;
  }

  /**
   * Multiply each coefficient and rhs by scale to get a new inequality.
   * If the scale is negative, the sense of the inequality reverses.
   * @param scale factor to scale coefficients and rhs by.
   * @precondition scale must not contain 0, unless the sense is equality.
   */
  public Inequality scaledIneq(Interval scale) {
    //1. check precondition.
    BigDecimal lower = scale.getCenter().subtract(scale.getRadius());
    BigDecimal upper = scale.getCenter().add(scale.getRadius());
    util.Eiffel.precondition(sense == CpxProblem.Equal || lower.signum()>=0 || upper.signum()<=0);
    //2. scale coefficients.
    Inequality ineq = new Inequality();
    for (Iterator I = this.getVariableIterator();I.hasNext();/*--*/){
      Variable V = (Variable) I.next();
      Interval coeff = this.getCoefficient(V);
      ineq.addCoeff(V,coeff.multiply(scale));
    }
    ineq.setRHS(this.rhs.multiply(scale));
    //3. set sense.
    ineq.setSense(this.sense);
    if (upper.signum()<=0) { // reverse inequality.
      if (this.sense == CpxProblem.LessEqual)
        ineq.setSense(CpxProblem.GreaterEqual);
      else if (this.sense == CpxProblem.GreaterEqual)
        ineq.setSense(CpxProblem.LessEqual);
    }
    return ineq;
  }

  /**
   * Multiply each coefficient and rhs by scale to get a new inequality.
   * If the scale is negative, the sense of the inequality reverses.
   * @param scale factor to scale coefficients and rhs by.
   */
  public Inequality scaledIneq(BigDecimal scale) {
    //2. scale coefficients.
    Inequality ineq = new Inequality();
    Variable V;
    Interval coeff;
    for (Iterator I = this.getVariableIterator();I.hasNext();/*--*/){
      V = (Variable) I.next();
      coeff = this.getCoefficient(V);
      ineq.addCoeff(V,coeff.multiply(scale));
    }
    ineq.setRHS(this.rhs.multiply(scale));
    //3. set sense.
    ineq.setSense(this.sense);
    if (scale.signum()<=0) { // reverse inequality.
      if (this.sense == CpxProblem.LessEqual)
        ineq.setSense(CpxProblem.GreaterEqual);
      else if (this.sense == CpxProblem.GreaterEqual)
        ineq.setSense(CpxProblem.LessEqual);
    }
    return ineq;
  }

  public char getSense() {
    return sense;
  }

  public String toString() {
    StringBuffer buf = new StringBuffer();
    /*
    for (Enumeration E=this.getVariableEnumeration();E.hasMoreElements();){
      Variable V = (Variable)E.nextElement();
      Interval coeff = this.getCoefficient(V);
      buf.append("["+coeff.toString()+" * "+V.toString()+ " ]  ");
    }
    */
    buf.append(super.toString());
    buf.append(' ');
    switch (this.getSense()) {
      case CpxProblem.GreaterEqual : buf.append('>'); break;
      case CpxProblem.LessEqual : buf.append('<'); break;
      case CpxProblem.Equal : buf.append('='); break;
      default : buf.append("?^?");
    }
    buf.append(" "+this.getRHS().toString());
    return buf.toString();
  }

  public void setRHS(Interval rhs) {
    this.rhs = rhs;
  }

  public Interval getRHS() {
    return rhs;
  }


  private Inequality() {
    super();
  }

  public Inequality(Object name) {
    super();
    this.name = name;
  }

  public static class TestCode extends UnitTest {
    public void test() {
      Inequality B = new Inequality();
    }
  }


}