package linprog;
import ilog.cplex.example.CpxProblem;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 * Company:
 * @author Thomas C. Hales
 * @version 2001.0
 */
import java.util.*;
import java.math.BigDecimal;

/**
 * This class generates inequalities needed in linear programming optimizations
 * of corrected volumes.  It is used by GraphLP to generate Inequalities.
 *
 * The InequalityTemplate class is used for general linear inequalities that
 * can be listed in a file.
 *
 * The inequalities here depend on convexity arguments of the underlying functions,
 * and the slopes and intercepts depend on bounds on the arguments, determined at runtime.
 *
 * This class also gives the assembly inequalities:
 *  The dihedral angles around each vertex sum to 2Pi.
 */

public class InequalityFactory {

final static private Parameters P = ParameterFactory.get();
final static private Variable threshold = Parameters.threshold;

private InequalityFactory() {
}

/**
 * Helper function.
 * Call this only after the RHS has been set.
 * It has a side-effect, modifying the input parameter so that it has a threshold variable.
 */
private static void addThreshold(Inequality ineq) {
  ineq.addCoeff(threshold,ineq.getRHS());
}

public static Inequality lowerBound(Variable V, Interval lb) {
  Inequality ineq = new Inequality("lower bound "+V.toString());
  ineq.addCoeff(V,Parameters.I_one);
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.setRHS(lb);
  addThreshold(ineq);
  return ineq;
}

public static Inequality lowerBoundI(Iterator varI, Interval lb) {
  Inequality ineq = new Inequality("lowerBoundI "+lb);
  while (varI.hasNext()) {
    Variable V = (Variable)varI.next();
    ineq.addCoeff(V,Parameters.I_one);
  }
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.setRHS(lb);
  addThreshold(ineq);
  return ineq;
}

public static Inequality upperBound(Variable V, Interval ub) {
  Inequality ineq = new Inequality("upper bound "+V.toString());
  ineq.addCoeff(V,Parameters.I_one);
  ineq.setSense(CpxProblem.LessEqual);
  ineq.setRHS(ub);
  addThreshold(ineq);
  return ineq;
}

public static Inequality upperBoundI(Iterator varI, Interval ub) {
  Inequality ineq = new Inequality("upper boundI "+ub);
  while (varI.hasNext()) {
    Variable V = (Variable)varI.next();
    ineq.addCoeff(V,Parameters.I_one);
  }
  ineq.setSense(CpxProblem.LessEqual);
  ineq.setRHS(ub);
  addThreshold(ineq);
  return ineq;
}

/**
 * Adih <= Amax dih
 */
public static Inequality upperBoundAdih(Adih adih,Angle dih,Interval Amax) {
  Inequality ineq = new Inequality("upper bound Adih");
  ineq.addCoeff(dih,Amax);
  ineq.addCoeff(adih,Parameters.I_mOne);
  ineq.setRHS(Parameters.I_zero);
  ineq.setSense(CpxProblem.GreaterEqual);
  addThreshold(ineq);
  return ineq;
}


private static Interval secant(IntervalFunction F,Interval hmin,Interval hmax) {
  Interval Fmax = F.value(hmax);
  Interval Fmin = F.value(hmin);
  Interval Fdiff = Fmax.subtract(Fmin);
  Interval hdiff = hmax.subtract(hmin);
  return Fdiff.divide(hdiff);
}


/**
 * The iterator contains all the Angles at the given Vertex.
 */
public static Inequality sum2Pi(Iterator AnglesAtVertex) {
  Inequality ineq = new Inequality("sum2Pi");
  ineq.setSense(CpxProblem.Equal);
  ineq.setRHS(Parameters.I_twopi);
  while (AnglesAtVertex.hasNext()) {
    Variable V = (Variable)AnglesAtVertex.next();
    ineq.addCoeff(V,Parameters.I_one);
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * -solF + sum dih = (n-2) pi.
 */
public static Inequality solid(Iterator AnglesOnFace, Variable F) {
  Inequality ineq = new Inequality("solid");
  int fcount = -2;
  ineq.setSense(CpxProblem.Equal);
  while (AnglesOnFace.hasNext()) {
    fcount++;
    Variable V = (Variable)AnglesOnFace.next();
    ineq.addCoeff(V,Parameters.I_one);
  }
  ineq.addCoeff(F,Parameters.I_mOne);
  Interval rhs = Parameters.I_pi.multiply(new BigDecimal(fcount));
  ineq.setRHS(rhs);
  addThreshold(ineq);
  return ineq;
}

public static Inequality sumIdentity(Iterator varInSum,Variable whatItEquals) {
  Inequality ineq = new Inequality("sumIdentity");
  ineq.setSense(CpxProblem.Equal);
  while (varInSum.hasNext()) {
    ineq.addCoeff((Variable)varInSum.next(),Parameters.I_one);
  }
  ineq.addCoeff(whatItEquals,Parameters.I_mOne);
  ineq.setRHS(P.I_zero);
  return ineq;
}

/**
 * sigma <= vor0 inequality.
 * vor0 = 4/3 solid - 4doct (phi0 sol - sum Adih + sum quoin)
 * so....
 *   4/3 solid - 4doct phi0 solid + sum 4 doct Adih - sum 4 doct quoin - sigma >= 0.
 */
public static Inequality vor0(Iterator AdihOnFace,Iterator QuoinOnFace,Solid sol,Sigma sig) {
  Interval four_3 = new Interval("4.0").divide(new Interval("3.0"));
  Parameters P = ParameterFactory.get();
  Interval phi0 = Trig.phi(P.trunc(),P.trunc());
  Interval doct4 = P.I_doct.multiply(new Interval("4.0"));
  Inequality ineq = new Inequality("vor0");
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.addCoeff(sig,P.I_mOne);
  ineq.setRHS(P.I_zero);
  ineq.addCoeff(sol,four_3);
  while (AdihOnFace.hasNext()) {
    Variable V = (Variable)AdihOnFace.next();
    ineq.addCoeff(V,doct4);
  }
  Interval mdoct4 = doct4.negate();
  while (QuoinOnFace.hasNext()) {
    Variable V = (Variable)QuoinOnFace.next();
    ineq.addCoeff(V,mdoct4);
  }
  return ineq;
}

/**
 * The iterator contains all the variables.
 * Make the inequality "sum var sense rhs."
 */
public static Inequality sumVar(Iterator varI, char sense, Interval rhs) {
  Inequality ineq = new Inequality("sumVar "+sense+" "+rhs);
  ineq.setSense(sense);
  ineq.setRHS(rhs);
  while (varI.hasNext()) {
    Variable V = (Variable)varI.next();
    ineq.addCoeff(V,Parameters.I_one);
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * sum sigma >= 8 pt.
 * varI contains all the sigmas.
 */
public static Inequality sigmaOver8pt(Iterator varI) {
  Inequality ineq = new Inequality("sigmaOver8pt");
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.setRHS(ParameterFactory.get().scoreTarget());
  while (varI.hasNext()) {
    Variable V = (Variable)varI.next();
    ineq.addCoeff(V,new Interval("1.0"));
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * sum_4 sigma >= 0.33 pt around type 4. SPIII, Group 4, 1.
 */
public static Inequality sum033(Iterator varI) {
  Interval pt033 = new Interval("0.33").multiply(P.I_pt);
  Inequality ineq = new Inequality("0.33");
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.setRHS(pt033);
  while (varI.hasNext()) {
    Variable V = (Variable) varI.next();
    ineq.addCoeff(V,P.I_one);
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * sum_5 sigma < sum_5 (-0.419351 sol + 0.2856354).  SPIII, Group 4, 3.
 */
public static Inequality sum5(Iterator varSigma,Iterator varSol) {
  Interval a = new Interval("0.419351");
  Interval b = new Interval("0.2856354").multiply(new BigDecimal("5.0"));
  Inequality ineq = new Inequality("sum5");
  ineq.setSense(CpxProblem.LessEqual);
  ineq.setRHS(b);
  while (varSigma.hasNext()) {
    Variable V = (Variable) varSigma.next();
    ineq.addCoeff(V,P.I_one);
  }
  while (varSol.hasNext()) {
    Variable V = (Variable) varSol.next();
    ineq.addCoeff(V,a);
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * sum_r tau > 0.55 pt k
 * k <=4, the number of distinct vertices in string.
 */
public static Inequality SPIII_Lemma53(Iterator varSigma,Iterator varSolid, int k) {
  Interval pt055k = new Interval("0.55").multiply(P.I_pt).multiply(new BigDecimal((double)k));
  Interval tauC = new Interval("0.100444572");
  Inequality ineq = new Inequality("SPIII_Lemma53");
  ineq.setSense(CpxProblem.GreaterEqual);
  ineq.setRHS(pt055k);
  while (varSigma.hasNext()) {
    Variable V = (Variable) varSigma.next();
    ineq.addCoeff(V,P.I_mOne);
  }
  while (varSolid.hasNext()) {
    Variable V = (Variable) varSolid.next();
    ineq.addCoeff(V,tauC);
  }
  addThreshold(ineq);
  return ineq;
}

/**
 * sum_r sigma < r pt - 0.48 k pt.
 * k <=4, the number of distinct vertices in string.
 */
public static Inequality SPIII_Lemma53b(Iterator varSigma, int k) {
  Interval pt048k = new Interval("0.48").multiply(P.I_pt).multiply(new BigDecimal((double)k));
  Interval rhs_accum = pt048k.negate();
  Inequality ineq = new Inequality("SPIII_Lemma53b");
  ineq.setSense(CpxProblem.LessEqual);
  while (varSigma.hasNext()) {
    Variable V = (Variable) varSigma.next();
    ineq.addCoeff(V,P.I_one);
    rhs_accum = rhs_accum.add(P.I_pt);
  }
  ineq.setRHS(rhs_accum);
  addThreshold(ineq);
  return ineq;
}


}

interface IntervalFunction {
    Interval value(Interval X);
}