package linprog;
import graph.Face;
import java.util.*;
import java.math.BigDecimal;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * This class serves as bridge between a MetricGraph and a GraphLP.
 * A MetricGraph should be thought of as a large data structure that is more or less passive.
 * The GraphLP reads data from the MetricGraph indirectly through the MetricGraphVariables
 *  object.
 *
 * The MetricGraphVariables creates all the linear programming variables that
 * go along with a given graph.  It has many getter and setter methods to access those
 * variables.
 *
 * MetricGraphVariables also exports certain methods from MetricGraph that might
 * be needed by GraphLP.
 */

public class MetricGraphVariables {

  private MetricGraph MG;

  public MetricGraphVariables(MetricGraph MG) {
    this.MG= MG;
    this.BuildVariables();
  }

  public MetricGraph getMetricGraph() {
    return MG;
  }

  // internal data.
  private boolean initialized = false;

  private HashMap vMap  = new HashMap();  // graph.Vertex -> Vertex.
  private Function edgeMap = new Function(); // e(pair) -> Edge.
  private Function quoinMap = new Function(); // e(directed pair) -> Quoin.
  private HashMap faceMap = new HashMap(); // F -> Solid.
  private Function angleMap = new Function(); // (F,v) -> Angle.
  private Function adihMap = new Function(); // (F,v)-> Adih.
  private HashMap sigmaMap = new HashMap(); // F-> Sigma

  /**
   * Value iterators:
   */
  public Iterator getSigma() {
    return valueIt(sigmaMap);
  }
  public Iterator getAdih() {
    return valueIt(adihMap);
  }
  public Iterator getV() {
    return valueIt(vMap);
  }
  public Iterator getEdge() {
    return valueIt(edgeMap);
  }
  public Iterator getQuoin() {
    return valueIt(quoinMap);
  }
  public Iterator getQuoinKeys() {
    return quoinMap.iterator();
  }
  public Iterator getAngle() {
    return valueIt(angleMap);
  }
  public Iterator getSolid() {
    return valueIt(faceMap);
  }

  /**
   * getters...
   */
  public Vertex varV(graph.Vertex V) {
    return (Vertex)vMap.get(V);
  }
  public Solid varSolid(graph.Face F) {
    return (Solid)faceMap.get(F);
  }
  public Sigma varSigma(graph.Face F) {
    return (Sigma)sigmaMap.get(F);
  }
  /**
   * Takes an iterator key and returns the Vertex.
   */
  public graph.Vertex edge0(Object[] key) {
    return  (graph.Vertex) key[0];
  }
  public graph.Vertex edge1(Object[] key) {
    util.Eiffel.assert(key!=null && key.length>1);
    return (graph.Vertex) key[1];
  }
  public Edge varE(graph.Vertex V1, graph.Vertex V2) {
    return (Edge)edgeMap.get(MG.sortPair(V1,V2));
  }
  public Quoin varQ(graph.Vertex V1, graph.Vertex V2) {
    return (Quoin)quoinMap.get(new Object[]{V1,V2});
  }
  public Angle varAngle(graph.Face F, graph.Vertex V) {
    util.Eiffel.assert(F.next(V,0)==V);
    return (Angle)angleMap.get(new Object[]{F,V});
  }
  public Adih varAdih(graph.Face F, graph.Vertex V) {
    util.Eiffel.assert(F.next(V,0)==V);
    return (Adih)adihMap.get(new Object[]{F,V});
  }

  /**
   * convertst the set of values of a hashmap to an iterator.
   */
  private static Iterator valueIt(HashMap h) {
    return h.values().iterator();
  }
  private static Iterator valueIt(Function h) {
    return h.values().iterator();
  }

  /**
   * exports function from MetricGraph.
   */
  public Interval getNexcept() {
    return MG.Nexcept();
  }

  public BigDecimal getMinHeight(graph.Vertex V1) {
    return MG.getMinHeight(V1);
  }

  public BigDecimal getMaxHeight(graph.Vertex V1) {
    return MG.getMaxHeight(V1);
  }

  public Interval getMinEdge(graph.Vertex V1, graph.Vertex V2) {
    return MG.getMinEdge(V1,V2);
  }

  public Interval getMaxEdge(graph.Vertex V1, graph.Vertex V2) {
    return MG.getMaxEdge(V1,V2);
  }

  public Enumeration vertexEnumeration() {
    return MG.vertexEnumeration();
  }

  public Enumeration faceEnumeration() {
    return MG.faceEnumeration();
  }

    /**
   * Initializes all the instance variables.  It is called by the constructor.
   */
  private void BuildVariables() {
    if (initialized) return; // make the function call idempotent.
    initialized=true;
    int counter=0;
    //1. name the vertices
    HashMap vName =  new HashMap();
    for (Enumeration E=MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      Object V = E.nextElement();
      vName.put(V,new Integer(counter++));
    }
    //2. name the faces
    counter =0;
    HashMap fName = new HashMap();
    for (Enumeration E = MG.faceEnumeration();E.hasMoreElements();/*--*/) {
      Object F = E.nextElement();
      fName.put(F,new Integer(counter++));
    }
    //3. create vertex variables
    for (Enumeration E=MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex)E.nextElement();
      vMap.put(V,new Vertex("_"+vName.get(V)));
      Face F = V.getAny();
      for (int i=0;i<V.size();i++) {
        Face Fx = V.next(F,i);
        angleMap.add(new Object[]{Fx,V},new Angle("_F"+fName.get(Fx)+"_v"+vName.get(V)));
        adihMap.add(new Object[]{Fx,V},new Adih("_F"+fName.get(Fx)+"_v"+vName.get(V)));
      }
      for (int i=0;i<V.size();i++) {
        graph.Vertex W = V.next(F,i).next(V,1);
        util.Eiffel.assert(MG.isEdge(V,W));
        Object[] pair = MG.sortPair(V,W);
        Object[] dpair = new Object[]{V,W};
        if (edgeMap.get(pair)==null) {
            String pairName = "("+vName.get(pair[0])+","+vName.get(pair[1])+")";
            edgeMap.add(pair,new Edge(pairName));
        }
        String dpairName = "("+vName.get(dpair[0])+"->"+vName.get(dpair[1])+")";
        quoinMap.add(dpair,new Quoin(dpairName));
      }
    }
    //4. create face variables
    for (Enumeration E = MG.faceEnumeration();E.hasMoreElements();/*--*/) {
      graph.Face F = (Face)E.nextElement();
      faceMap.put(F,new Solid(""+fName.get(F)));
      sigmaMap.put(F,new Sigma(""+fName.get(F)));
    }
  }


}