package linprog;
import java.util.*;
import java.lang.StringBuffer;
import util.*;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * Let's make a simple parser for the following grammer.
 * Tokens
 *   :, +, -, <, >, =, [, ], ,, unsigneddecimalnumbers, symbols
 *   whitespace (tab and space and cr).
 *   comments: rest of line starting with '#'
 *   states:
 *    whitespace : swallow up
 *    comment : swallow to end of line.
 *    symbol : alphanumeric continue with a-z, _, 0-9,
 *    unsigneddecimal : 0-9 continue with 0-9, . `
 *    singletoken: read one.
 *
 * InequalityList := (domain3 (inequality)* )+
 *
 *  inequality := name <COLON> linCombo sense linCombo
 *  linCombo := (term)+
 *  term := number symbol
 *  sense:= <LESS> | <GREATER> | <EQUALS>
 *
 * domain3:= domainVar domain domain
 * domainVar := <LEFTBRACKET> symbol (<COMMA> symbol)* <RIGHTBRACKET>
 * number := (<PLUS>|<MINUS>)? unsigneddecimalnumber
 * domain := <LEFTBRACKET> number (<COMMA> number)* <RIGHTBRACKET>
 *
 * Example in the file test.data.
 *

 *
 */

public class ParseIneq {

  private PeekIterator E;
  private final String S;

  public ParseIneq(String S) {
    this.S = S;
    E = newIter();
  }

  private PeekIterator newIter() {
    return new PeekIterator(new Iterator() {
      int i=0;
      String buff = S;
      public boolean hasNext() {
        return (i < buff.length());
      }
      public Object next() {
        return new Character(S.charAt(i++));
      }
      public void remove() {
        throw new java.lang.UnsupportedOperationException();
      }
    });
  }

  private int token_state;
  public static final int whitespace =0;  // space, tab, cr.
  public static final int comment=1; // #,
  public static final int symbol=2; // a-zA-Z
  public static final int unsigneddecimal=3; // 0-9 .
  public static final int singlechar=4; // + - [ ] , < > = :

  private int updateState(int state, char c) {
    if (state==comment) {
      return (c=='\n' ? whitespace : comment);
    }
    if (state==symbol && alphaNumeric(c)) {
      return symbol;
    }
    return charType(c);
  }

  private int charType(char c) {
    if (c=='#') return comment;
    if (c>='0' && c<='9') return unsigneddecimal;
    if (c=='.' || c=='`') return unsigneddecimal;
    if (c>='a' && c<='z') return symbol;
    if (c>='A' && c<='Z') return symbol;
    if (c=='_') return symbol;
    if (c==':' || c=='+' || c=='-' || c=='[' || c==']' || c==',' || c=='<' || c=='>' || c=='=')
      return singlechar;
    if (c==' ' || c=='\t' || c=='\n')
      return whitespace;
    // no other chars are allowed, but we will be graceful about the errors,
    // and allow them to start a comment.
    return comment; // no other characters are alllowed.
  }

  private boolean alphaNumeric(char c) {
    if (c>='0' && c<='9') return true;
    if (c>='a' && c<='z') return true;
    if (c>='A' && c<='Z') return true;
    if (c=='_') return true;
    return false;
  }

  private char nextChar(PeekIterator I) {
    return ((Character)I.next()).charValue();
  }

  private char peekChar(PeekIterator I) {
    return ((Character)I.peek()).charValue();
  }

  private Token nextToken(PeekIterator I) {
    return ((Token)I.next());
  }

  private Token peekToken(PeekIterator I) {
    return  ((Token)I.peek());
  }

  private boolean headToken(PeekIterator I,char c) {
    Token T = peekToken(I);
    return (T.isSingle(c));
  }

  private Token scanToken(PeekIterator I) {
    StringBuffer buff = new StringBuffer();
    if (!I.hasNext())
      return null;
    int state = this.charType(peekChar(I));
    if (state==singlechar) {
      buff.append(nextChar(I));
      return new Token(buff.toString(),singlechar);
    }
    while (I.hasNext()) {
      char c = peekChar(I);
      if (updateState(state,c)!=state) {
        return new Token(buff.toString(),state);
      }
      buff.append(nextChar(I));
    }
    return new Token(buff.toString(),state);
  }

  public PeekIterator Tokenize() {
    ArrayList L = new ArrayList();
    Token T;
    // ignore comments and whitespace.
    while ((T=scanToken(E))!=null) {
      if (T.isSymbol() || T.isSingle() || T.isNumber())
        L.add(T);
    }
    return new PeekIterator(L.iterator());
  }




 /**
  * Production rules
  */
  public Node parseInequalityList(PeekIterator I) {
    Node N = new Node(Node.INEQUALITYLIST);
    while (I.hasNext()) {
      if (headToken(I,'['))
        N.addChild(parseDomain3(I));
      else
        N.addChild(parseInequality(I));
    }
    return N;
  }

  Node parseDomain3(PeekIterator I) {
    Node N = new Node(Node.DOMAIN3);
    N.addChild(parseDomainVar(I));
    N.addChild(parseDomain(I));
    N.addChild(parseDomain(I));
    return N;
  }

  private Node parseDomain(PeekIterator I) {
    Node N = new Node(Node.DOMAIN);
    parseSingle(I,'[');
    N.addChild(parseNumber(I));
    while (I.hasNext() && headToken(I,',')) {
      parseSingle(I,',');
      N.addChild(parseNumber(I));
    }
    parseSingle(I,']');
    return N;
  }

  private Node parseDomainVar(PeekIterator I) {
    Node N = new Node(Node.DOMAINVAR);
    parseSingle(I,'[');
    N.addChild(parseToken(I,Node.SYMBOL));
    while (I.hasNext() && headToken(I,',')) {
      parseSingle(I,',');
      N.addChild(parseToken(I,Node.SYMBOL));
    }
    parseSingle(I,']');
    return N;
  }

 private Node parseInequality(PeekIterator I) {
  Node N = new Node(Node.INEQUALITY);
  N.addChild(parseToken(I,Node.SYMBOL)); // name;
  parseSingle(I,':');
  N.addChild(parseLinCombo(I));
  N.addChild(parseToken(I,Node.SENSE));
  N.addChild(parseLinCombo(I));
  return N;
 }

 private String dumpIterator(Iterator I) {
  StringBuffer b = new StringBuffer();
  while (I.hasNext()) {
    b.append(I.next());
  }
  return b.toString();
 }

 private void parseSingle(PeekIterator I,char c) {
  util.Eiffel.assert(I.hasNext());
  Token T = nextToken(I);
  String errMsg="";
  if (!T.isSingle(c)) {
    errMsg="bad char("+c+")***("+T.toString()+")***("+dumpIterator(I)+")";
  }
  util.Eiffel.assert(T.isSingle(c),errMsg);
 }

 private Node parseToken(PeekIterator I,int nodeType) {
  util.Eiffel.assert(I.hasNext());
  Token T = nextToken(I);
  Node N = new Node(nodeType);
  N.addChild(T.getString());
  return N;
 }

 private Node parseLinCombo(PeekIterator I) {
  Node N = new Node(Node.LINCOMBO);
  while (I.hasNext() &&
      (peekToken(I).isNumber() || peekToken(I).isSingle('+') || peekToken(I).isSingle('-'))
      )
    N.addChild(parseTerm(I));
  return N;
 }

 private Node parseTerm(PeekIterator I) {
  Node N = new Node(Node.TERM);
  N.addChild(parseNumber(I));
  N.addChild(parseToken(I,Node.SYMBOL));
  return N;
 }

 private Node parseNumber(PeekIterator I) {
  Node N = new Node(Node.NUMBER);
  int sign = 1;
  if (headToken(I,'+'))
    nextToken(I);
  else if (headToken(I,'-')) {
    nextToken(I);
    sign=-1;
  }
  Token T = nextToken(I);
  util.Eiffel.assert(T.isNumber());
  String S = T.getString();
  Interval A = StringToInterval(S);
  if (sign<0) A = A.negate();
  N.addChild(A);
  return N;
 }

 private Interval StringToInterval(String S) {
  int index = S.indexOf('`');
  if (index<0) return new Interval(S);
  String center = S.substring(0,index);
  String radius = S.substring(index+2);
  return new Interval(center,radius);
 }

  public static class TestCode extends UnitTest {
    public void testToken() {
      StringBuffer B = new StringBuffer();
      ParseIneq PI = new ParseIneq("This  y1 +3.4`0 ><\t\n = #oioi\nand");
      PeekIterator I = PI.Tokenize();
      while (I.hasNext()) {
        B.append(PI.nextToken(I).toString());
      }
      assert(B.toString().equals("{SYMBOL,This}{SYMBOL,y1}{CHAR,+}{UNSIGNED,3.4`0}{CHAR,>}"+
        "{CHAR,<}{CHAR,=}{SYMBOL,and}"));
    }

    public void testStringToInterval() {
      String S = "345.00`0.001";
      assert(new ParseIneq("").StringToInterval(S).equals(new Interval("345.0","0.001")));
    }

  }
}



/***
 * NEW CLASS TOKEN
 */
class Token {
  private String S;
  public String getString() {
    return S;
  }
  private int type;
  private String tag() {
    switch (type) {
      case ParseIneq.whitespace : return "WHITE";
      case ParseIneq.comment : return "COMMENT";
      case ParseIneq.symbol: return "SYMBOL";
      case ParseIneq.unsigneddecimal: return "UNSIGNED";
      case ParseIneq.singlechar: return "CHAR";
      default: return "UNKNOWN";
    }
  }

  Token(String S,int type) {
    this.type = type;
    this.S = S;
  }
  boolean isSymbol() {
    return type==ParseIneq.symbol;
  }
  boolean isNumber() {
    return type==ParseIneq.unsigneddecimal;
  }
  boolean isSingle() {
    return type==ParseIneq.singlechar;
  }
  boolean isSingle(char c) {
    return (type==ParseIneq.singlechar&& c==S.charAt(0));
  }

  public String toString() {
    return '{'+tag()+','+S+'}';
  }

}


