package linprog;

import ilog.cplex.example.CpxProblem;
import ilog.cplex.example.CpxException;
import ilog.cplex.example.CpxSparseVector;


/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 * Company:
 * @author Thomas C. Hales
 * @version 2001.0
 */

import java.util.*;
import util.UnitTest;
import util.Eiffel;
import java.math.BigDecimal;
import graph.CalendarExtension;

/**
 * contains the data for one linear program.
 * Contains a list of variables, each variable has an internal number inside this class.
 * Contains an Objective object, and a collection of inequalities.
 * <p>
 * Variables are handled with addVariable and addVariableByName
 * <p>
 * Objective with setObjective.
 * <p>
 * Inequalities with addInequality, removeInequality.
 * <p>
 * optimum is a numerically rigorous bound.
 * <p>
 * Threshold:  A dummy variable called threshold is introduced into a modified
 * problem in a way that insures a
 * feasible bounded solution.
 * It is the responsibility of the caller to create the threshold-modified
 * inequalitie and objective functions.
 * <UL>
 * <LI> threshold has variable bounds [0,1].
 * <LI>Threshold appears in the objective as  + C*threshold
 * <LI>Threshold appears in each inequality as + getRHS()*threshold,
 *    (excepting bounding inequalities on threshold itself).
 * <LI>The modified solution has a feasible solution with objective C, (at x=0, threshold=1).
 * <LI>If the original solution has objective beyond C, the optimal value is unchanged.
 * </UL>
 */
public class LPdata {

  //private Hashtable variableLookup = new Hashtable();
  private Hashtable variableNumberLookup = new Hashtable(); // Variable -> Integer
  private Hashtable numberToVariable = new Hashtable(); // Integer -> Variable (inverse function)
  private HashMap solutionVariable = null; // Variable -> Double
  private static int globalcounter=0;

  public static int getGlobalCounter() {
    return globalcounter;
  }

  public HashMap getSolutionVariable() throws CpxException {
    //this.optimum(true);
    return solutionVariable;
  }

  private int count = 0;
  private static final boolean tracing = true;

  private Objective obj;
  private ArrayList ineqList = new ArrayList();

  public Enumeration variableEnumeration() {
    return variableNumberLookup.keys();
  }

  public String toString() {
    StringBuffer b = new StringBuffer();
    if (obj==null)
      b.append("NO OBJECTIVE");
    else
      b.append(obj);
    b.append("\n");
    for (int i=0;i<ineqList.size();i++) {
      b.append(ineqList.get(i).toString());
      b.append('\n');
    }
    return b.toString();
  }

  /**
   * checks that each inequality has been given a threshold variable,
   * with coeff equal to the RHS.
   * The only exception should be the lower bound 'threshold >=0'
   * This doesn't do exhaustive checking, but it should catch any major errors.
   */
  public void verifyThreshold(Variable threshold) {
    for (int i=0;i<ineqList.size();i++) {
      Inequality ineq = (Inequality)ineqList.get(i);
      Interval tCoeff = ineq.getCoefficient(threshold);
      Interval rhs = ineq.getRHS();
      //1. if tCoeff is null, rhs should be 0.
      if (tCoeff==null) {
        util.Eiffel.assert(rhs.max().signum()==0 && rhs.min().signum()==0,ineq.toString());
        continue;
      }
      boolean equal = ( rhs.max().subtract(tCoeff.max()).signum()==0
                        && rhs.min().subtract(tCoeff.min()).signum()==0);
      boolean greater = (ineq.getSense()==CpxProblem.GreaterEqual
                        && tCoeff.min().subtract(rhs.max()).signum()>0);
      boolean less = (ineq.getSense()==CpxProblem.LessEqual
                        && tCoeff.max().subtract(rhs.min()).signum()<0);
      util.Eiffel.assert((equal || greater || less),ineq.toString());
      /**
      util.Eiffel.assert(ineq.getCoefficient(threshold).subtract(
        ineq.getRHS()).getCenter().signum()==0
        ||
        ineq.getRHS().getCenter().signum()==0
        );
      **/
    }
  }

  /**
   * sets the objective function for the linear programming problem to obj.
   */
  public void setObjective(Objective obj) {
    this.obj = obj;
  }

  /**
   * Adds an inequality to the set of inequalities in the linear programming problem.
   */
  public void addInequality(Inequality ineq) {
    if (ineq!=null)
      ineqList.add(ineq);
  }

  /**
   * Removes an inequality from the set of inequalities in the LP.
   * This works at the level of objects, so an equivalent inequality that
   * is a different java object will not be removed.
   */
  public void removeInequality(Inequality ineq) {
    ineqList.remove(ineq);
  }

  /**
   * adds a variable to the LP, including two new inequalities corresponding
   * to the variable's upper and lower bounds.
   * It is an error to call optimum without declaring all the variables with
   * this method.
   * The threshold variable appears in all inequalities, including those for
   * variable bounds.
   */
  public void addVariable(Variable V, Variable threshold) {
    if (!variableNumberLookup.containsValue(V)) {
      variableNumberLookup.put(V,new Integer(count));
      numberToVariable.put(new Integer(count),V);
      count++;
      //create two new inequalities corresponding to upper and lower bounds.
      BigDecimal zero = new BigDecimal("0.0");
      Inequality lowerBound = new Inequality("var lower bound "+V.toString());
      lowerBound.setSense(CpxProblem.GreaterEqual);
      lowerBound.setRHS(new Interval(V.getLowerBound(),zero));
      lowerBound.addCoeff(V, new  Interval("1.0"));
      if (threshold!=null)
        lowerBound.addCoeff(threshold, lowerBound.getRHS());
      ineqList.add(lowerBound);
      // upper
      Inequality upperBound = new Inequality("var upper bound "+V.toString());
      upperBound.setSense(CpxProblem.LessEqual);
      upperBound.setRHS(new Interval(V.getUpperBound(),zero));
      upperBound.addCoeff(V, new Interval("1.0"));
      if (threshold!=null)
        upperBound.addCoeff(threshold, upperBound.getRHS());
      ineqList.add(upperBound);
    }
    //variableLookup.put(V.getName(),V);
  }

  private Variable getVariableByNumber(int number) {
    return (Variable)numberToVariable.get(new Integer(number));
  }

  private int getNumberOfVariable(Variable V) {
    Integer a = (Integer)variableNumberLookup.get(V);
    util.Eiffel.assert(a!=null,"var not found "+V.toString());
    return a.intValue();
  }

  private static final double epsilon = 1.0e-14; // treat as 0 anything under this.
  /**
   * Sum all inequalities with weights from dual solution.
   * @param cplex A CpxProblem that has already been optimized.
   * @param scaledSense sense of return value. Must be <= or >=, and not simply =;
   */
  private Inequality InequalitySum(CpxProblem cplex, char sense) throws CpxException {
    util.Eiffel.assert(sense==CpxProblem.LessEqual || sense==CpxProblem.GreaterEqual);
    Inequality sum = new Inequality("**temp**");
    sum.setSense(sense);
    BigDecimal zero = new BigDecimal("0.0");
    sum.setRHS(new Interval("0.0"));
    int nrows = cplex.getNRows();
    util.Eiffel.assert(nrows== ineqList.size());
    double[] pi    = new double[nrows];
    cplex.getPi(pi);
    //if (tracing)
      //System.out.println("just got cplex data. now do interval stuff");
    BigDecimal scale;
    Inequality B;
    BigDecimal mOne = new BigDecimal(-1.0);
    java.util.Set S = new java.util.HashSet();
    for (int i=0;i<nrows;i++) {
      if (Math.abs(pi[i])< epsilon) // this affects the quality, but not the rigor of the solution.
        continue;
      scale = new BigDecimal(pi[i]);
      Inequality ineq = (Inequality)ineqList.get(i);
      if (tracing)
        S.add(ineq.getName());
      B = ((Inequality)ineqList.get(i)).scaledIneq(scale);
      if (B.getSense()!=ilog.cplex.example.CpxProblem.Equal &&
            B.getSense()!= sense) {
            B = B.scaledIneq(mOne);
      }
      sum.augmentBy(B);
    }
    /*
    for (Iterator I = S.iterator();I.hasNext();) {
      if (tracing)
        System.out.println(I.next());
    }
    */
    return sum;
  }

  /**
   * bounds on variables are entered as constraints to facilitate dual checking.
   * So all variables are entered as free.  This method is based on examples
   * from the javaCplex distribution.
   */
  private void populatebyrow (CpxProblem cplex) throws CpxException
   {
      //1. get number of rows and columns.
      int ncols = variableNumberLookup.size();
      int nrows = ineqList.size();
      util.Eiffel.assert(variableNumberLookup.size()==count);
      //util.Eiffel.assert(count==variableLookup.size());
      util.Eiffel.assert(count==numberToVariable.size());
      CpxSparseVector[] cols = new CpxSparseVector[ncols];

      //2. set objective
      cplex.setObjSense(obj.getSense());
      double       objective[]  = new double[ncols];
      for (Iterator I = obj.getVariableIterator();I.hasNext();/*--*/) {
        Variable V = (Variable) I.next();
        int pos = getNumberOfVariable(V);
        util.Eiffel.assert(pos>=0);
        util.Eiffel.precondition(pos< ncols,"pos col = "+pos+" "+ncols);
        objective[pos] = obj.getCoefficient(V).getCenter().doubleValue();
      }

      //3. set bounds (all free variables)
      double lb[] = new double[ncols];
      double ub[] = new double[ncols];
      char type[]= new char[ncols];
      for (int i=0;i<lb.length;i++) {
        lb[i] = -CpxProblem.Infinity;
        ub[i] = CpxProblem.Infinity;
        type[i]=CpxProblem.Continuous;
      }
      cplex.addCols(objective, cols, lb, ub, type);

      //4. add rows for original constraints
      CpxSparseVector[] rows  = new CpxSparseVector[nrows];
      char[] sense = new char[nrows];
      double[] rhs = new double[nrows];
      for (int i=0;i<ineqList.size();i++) {
        Inequality ineq = (Inequality)ineqList.get(i);
        sense[i]= ineq.getSense();
        rhs[i]= ineq.getRHS().getCenter().doubleValue();
        rows[i] = new CpxSparseVector();
        for (Iterator I = ineq.getVariableIterator();I.hasNext();/*--*/){
          Variable V = (Variable) I.next();
          int pos = getNumberOfVariable(V);
          rows[i].add(pos, ineq.getCoefficient(V).getCenter().doubleValue());
        }
      }

      cplex.addRows(rows, sense, rhs);
   }

  public BigDecimal optimum() throws CpxException {
    return optimum(true);
  }

   /**
    * Creates a new CpxProblem object, optimizes, returns rigorous bound.
    */
  private BigDecimal optimum(boolean setSolutionVariable) throws CpxException {
    //1. set cplex problem data
    CpxProblem cplex = new CpxProblem();
    cplex.setIntParam(1035, 1);
    populatebyrow(cplex);

    //2. do cplex optimization
    cplex.dualopt();
    if (tracing)
      System.out.println("["+(globalcounter++)+"] objective is informally = "+cplex.getObj());

    //3. sum inequalities using dual variable weights.
    char sense = (obj.getSense() == CpxProblem.Maximize
      ? CpxProblem.LessEqual
      : CpxProblem.GreaterEqual
     );
    Inequality sum = InequalitySum(cplex,sense);

    //4. look at solution
    if (setSolutionVariable) {
      solutionVariable = new HashMap();
      int ncols = variableNumberLookup.size();
      double[] xx    = new double[ncols];
      cplex.getX(xx);
      for (int i=0;i<ncols;i++) {
        Variable V = (Variable)numberToVariable.get(new Integer(i));
        util.Eiffel.assert(V!=null);
        solutionVariable.put(V,new Double(xx[i]));
      }
    }
    else {
      solutionVariable=null;
    }

    //5. dispose of the cplex instance
      try {
        cplex.dispose();
      }
      catch (Throwable t) {
        System.out.println("wait 1 sec");
      }
    ;//throw new CpxException(-1,"license not released");
    sum = sum.diminishBy(obj);
    return sum.gap();

  }

  public static void main(String[] args) {
    try {
    LPdata data = new LPdata();
    Variable threshold = new Variable(new BigDecimal(0.0),new BigDecimal(1.0));
    Variable x1 = new Variable(new BigDecimal(0.0),new BigDecimal(40.0));
    Variable x2 = new Variable(new BigDecimal(0.0),new BigDecimal(200.0));
    Variable x3 = new Variable(new BigDecimal(0.0),new BigDecimal(200.0));
    data.addVariable(threshold, null);
    data.addVariable(x1, threshold);
    data.addVariable(x2, threshold);
    data.addVariable(x3, threshold);
    Objective obj = new Objective();
    obj.setSense(CpxProblem.Maximize);
    obj.addCoeff(x1,new Interval("1.0"));
    obj.addCoeff(x2,new Interval("2.0"));
    obj.addCoeff(x3,new Interval("3.0"));
    obj.addCoeff(threshold,new Interval("202.0"));

    data.setObjective(obj);

    Inequality ineq1 = new Inequality("**test case");
    ineq1.setSense(CpxProblem.LessEqual);
    ineq1.setRHS(new Interval("20.0"));
    ineq1.addCoeff(x1,new Interval("-1.0"));
    ineq1.addCoeff(x2,new Interval("1.0"));
    ineq1.addCoeff(x3,new Interval("1.0"));
    ineq1.addCoeff(threshold,ineq1.getRHS());
    data.addInequality(ineq1);

    Inequality ineq2 = new Inequality("test case2");
    ineq2.setSense(CpxProblem.LessEqual);
    ineq2.setRHS(new Interval("30.0"));
    ineq2.addCoeff(x1,new Interval("1.0"));
    ineq2.addCoeff(x2,new Interval("-3.0"));
    ineq2.addCoeff(x3,new Interval("1.0"));
    ineq2.addCoeff(threshold,ineq2.getRHS());
    data.addInequality(ineq2);

    BigDecimal answer = data.optimum();
    System.out.println("optimum is " + answer.toString());
    //System.out.println(data.toString());
    data.verifyThreshold(threshold);
    }
    catch (CpxException e) {
      System.out.println("exception encountered");
    }

  }





}