package linprog;
import graph.Graph;
import graph.Face;
import java.util.*;
import java.math.BigDecimal;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

public class MetricGraph extends graph.Graph implements Cloneable {

  private static final Parameters P = ParameterFactory.get();

  /**
   * We need to keep track of all the keys to HashTables to facilitate cloning.
   * This is an aspect in the sense of AOP.  We need to check each time an entry
   * is added to any of the tables to make sure the key is in keySet.
   */
  private void preKeyAssert(Object[] obj) {
    for (int i=0;i<obj.length;i++) {
      util.Eiffel.assert(validPreKey(obj[i]));
    }
  }
  private boolean validPreKey(Object obj) {
    if (obj instanceof graph.Face)
      return this.contains((graph.Face) obj);
    if (obj instanceof graph.Vertex)
      return this.contains((graph.Vertex) obj);
    return false;
  }

  private void keyAssert(Object obj) {
    util.Eiffel.assert(validPreKey(obj));
  }

  private void keyAssertArray(Object[] obj) {
    if (obj==null)
      return;
    for (int i=0;i<obj.length;i++) {
      keyAssert(obj[i]);
    }
  }

  /**
   * originalFaceProfile[i]== # of faces in original graph that are i-gons.
   * precondition: N>=0;
   */
  private static final int largestNgon=10;
  private int[] originalFaceProfile = new int[largestNgon];

  /**
   * sum_{excep faces f} (N(f)+1)/3.
   */
  public Interval Nexcept() {
    int sum = 0;
    for (int i=5;i<originalFaceProfile.length;i++) {
      sum += originalFaceProfile[i]*(i+1);
    }
    // convert sum/3 to interval.
    Interval A = new Interval(new BigDecimal(sum),new BigDecimal(0));
    return A.divide(new Interval("3.0"));
  }

  public int getNumberNgons(int N) {
    util.Eiffel.assert(N>=0);
    if (N>=largestNgon)
      return 0;
    return originalFaceProfile[N];
  }

  /**
   * Helper to look up an entire array in a hashmap
   */
  private Object[] dict(Object[] obj, HashMap M) {
    Object[] lookup = new Object[obj.length];
    for (int i=0;i<obj.length;i++) {
      lookup[i]= M.get(obj[i]);
      if (lookup[i]==null)
        return null;
    }
    return lookup;
  }

  /**
   * Object[] -> Set of Strings;
   */
  private Function properties  = new Function();
  final private static String FATQUAD="FATQUAD"; // {Face} (both diags > 2sqrt2)
  final private static String SMALLQRTET="SMALLQRTET"; // {Face}  // if y4+y5+y6<=6.25.
  final private static String BIGQRTET="BIGQRTET"; // {Face} // if y4+y5+y6>=6.25;
  final private static String UPRIGHT="UPRIGHT";// {Vertex,Face} // if upright quarter;
  final private static String FLATQUARTER="FLATQUARTER"; // {Vertex,Face} // if flat quarter in a quad.

  public boolean isQRTET(Face F) {
    if (F.size()!=3) return false;
    graph.Vertex V = F.getAny();
    for (int i=0;i<3;i++) {
      graph.Vertex W = F.next(V,i);
      if (isUPRIGHT(W,F)) return false;
      if (isFLATQUARTER(W,F)) return false;
    }
    return true;
  }

  public boolean isORIGINALQUAD(Face F) {
    if (F.size()!=4) return false;
    if (isFATQUAD(F)) return false;
    return true;
  }

  public void setSmallQRTET(Face F) {
    util.Eiffel.assert(isQRTET(F));
    setProperty(new Object[]{F},SMALLQRTET);
  }

  public boolean isSmallQRTET(Face F) {
    return hasProperty(new Object[] {F},SMALLQRTET);
  }

  public void setBIGQRTET(Face F) {
    util.Eiffel.assert(isQRTET(F));
    setProperty(new Object[]{F},BIGQRTET);
  }

  public boolean isBIGQRTET(Face F) {
    return hasProperty(new Object[] {F},BIGQRTET);
  }

  public void setFLATQUARTER(graph.Vertex V, Face F) {
    util.Eiffel.assert(F.size()==3);
    util.Eiffel.assert(V.next(F,0)==F);
    setProperty(new Object[]{V,F},FLATQUARTER);
  }

  public boolean isFLATQUARTER(graph.Vertex V, Face F) {
    Object[] obj = new Object[] {V,F};
    return hasProperty(obj,FLATQUARTER);
  }

  public void setUPRIGHT(graph.Vertex V, Face F) {
    util.Eiffel.assert(F.size()==3);
    util.Eiffel.assert(V.next(F,0)==F);
    setProperty(new Object[]{V,F},UPRIGHT);
  }

  public boolean isUPRIGHT(graph.Vertex V, Face F) {
    Object obj[] = new Object[] {V,F};
    return hasProperty(obj,UPRIGHT);
  }

  public void setFATQUAD(Face F) {
    util.Eiffel.assert(F.size()==4);
    setProperty(new Object[]{F},FATQUAD);
  }

  public boolean isFATQUAD(Face F) {
    return hasProperty(new Object[]{F},FATQUAD);
  }



  private boolean hasProperty(Object[] obj,String name) {
    keyAssertArray(obj);
    Set M = (Set)properties.get(obj);
    if (M==null) return false;
    if (M.contains(name)) return true;
    return false;
  }

  public graph.Face getOriginalQuad() {
    if (originalQuadExhausted)
      return null;
    for (Enumeration E = this.faceEnumeration();E.hasMoreElements();/*--*/) {
      graph.Face F = (graph.Face)E.nextElement();
      if (F.size()==4   && !this.isFATQUAD(F))
        return F;
    }
    originalQuadExhausted=true;
    return null;
  }

  public boolean mightHaveOriginalQuad() {
    return !originalQuadExhausted;
  }
  private boolean originalQuadExhausted=false;

  public void removeProperty(Object[] obj,String name) {
    keyAssertArray(obj);
    Set M = (Set)properties.get(obj);
    if (M!=null) {
      M.remove(name);
    }
  }

  private void setProperty(Object[] obj, String name) {
    keyAssertArray(obj);
    Set M = (Set)properties.get(obj);
    if (M==null) {
      M = new HashSet();
      properties.add(obj,M);
    }
    M.add(name);
  }

  /**
  public void setProperty(Object obj, String name) {
    setProperty(new Object[] {obj},name);
  }
  **/

  /**
   * Vertex -> BigDecimal (minimum height of the vertex).
   */
  private HashMap minHeight = new HashMap();

  public void setMinHeight(graph.Vertex V,BigDecimal A) {
    keyAssert(V);
    minHeight.put(V,A);
  }

  public BigDecimal getMinHeight(graph.Vertex V) {
    keyAssert(V);
    return (BigDecimal)minHeight.get(V);
  }

  /**
   * Vertex -> BigDecimal (maximum height of the vertex).
   */
   private HashMap maxHeight = new HashMap();

   public void setMaxHeight(graph.Vertex V, BigDecimal A) {
    keyAssert(V);
    maxHeight.put(V,A);
   }

   public BigDecimal getMaxHeight(graph.Vertex V) {
    keyAssert(V);
    return (BigDecimal) maxHeight.get(V);
   }

   /**
    * Each edge is an unordered pair of vertices (v1,v2)
    */

    public boolean isEdge(graph.Vertex V1, graph.Vertex V2) {
      graph.Face F = V1.getAny();
      for (int i=0;i<V1.size();i++) {
        if (V1.next(F,i).next(V1,1)==V2)
          return true;
      }
      return false;
    }

  /**
   * sorted(V1,V2) -> Interval
   */
  private Function minEdge = new Function();
  private Function maxEdge = new Function();

  /**
   * Each edge has a minLength and maxLength;
   * precondition: (V1,V2) is an edge
   */
  public Interval getMinEdge(graph.Vertex V1, graph.Vertex V2) {
    util.Eiffel.assert(isEdge(V1,V2));
    return (Interval) minEdge.get(sortPair(V1,V2));
  }

  public Interval getMaxEdge(graph.Vertex V1, graph.Vertex V2) {
    util.Eiffel.assert(isEdge(V1,V2));
    return (Interval) maxEdge.get(sortPair(V1,V2));
  }

  public void setMinEdge(graph.Vertex V1, graph.Vertex V2, Interval A) {
    keyAssert(V1); keyAssert(V2);
    util.Eiffel.assert(isEdge(V1,V2));
    minEdge.add(sortPair(V1,V2),A);
  }

  public void setMaxEdge(graph.Vertex V1, graph.Vertex V2, Interval A) {
    keyAssert(V1); keyAssert(V2);
    util.Eiffel.assert(isEdge(V1,V2));
    maxEdge.add(sortPair(V1,V2),A);
  }

  protected static Object[] sortPair(graph.Vertex V1, graph.Vertex V2) {
    if (V1.hashCode() > V2.hashCode()) {
      return new Object[] {V2,V1};
    }
    return new Object[] {V1,V2};
  }

  /**
   * Constructor.
   * Shallow copy of G is made, so that Vertices, Faces, etc. are
   * shared between G and the MetricGraph.
   */
  public MetricGraph(Graph G) {
    super(G); // shallow clone of data.
    BigDecimal two = new BigDecimal("2.0");
    Interval I_two = new Interval(two);
    for (Enumeration E = G.vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex)E.nextElement();
      setMinHeight(V,two);
      setMaxHeight(V,P.eM().max());
    }
    checkDistinctVHash();
    for (Enumeration E = G.faceEnumeration();E.hasMoreElements();/*--*/) {
      graph.Face F = (graph.Face)E.nextElement();
      util.Eiffel.assert(this.contains(F));
      originalFaceProfile[F.size()]++;
      graph.Vertex V = F.getAny();
      for (int i=0;i<F.size();i++) {
        graph.Vertex V1 = F.next(V,i), V2 = F.next(V,i+1);
        setMaxEdge(V1,V2,P.eM());
        setMinEdge(V1,V2,I_two);
      }
    }
  }

  private void checkDistinctVHash() {
    Set S = new HashSet();
    for (Enumeration E = vertexEnumeration();E.hasMoreElements();/*--*/) {
      graph.Vertex V = (graph.Vertex)E.nextElement();
      Integer K = new Integer(V.hashCode());
      util.Eiffel.assert(!S.contains(K));
      S.add(K);
    }
  }


  /**
   * create a new metric graph with one more triangle than the current one.
   * two flat quarters are produced and labeled as such.
   * {F.next(V,-1),F.next(V,0),F.next(V,1)};
   */
  public MetricGraph divideQuad(graph.Face F, graph.Vertex V,
      Interval minL, Interval maxL) {
    util.Eiffel.assert(this.contains(F));
    util.Eiffel.assert(this.contains(V));
    util.Eiffel.assert(F.next(V,0)==V);
    util.Eiffel.assert(F.size()==4);
    util.Eiffel.assert(F.isFinal());
    F.setFinal(false);
    graph.Vertex ram1 = F.next(V,1);
    graph.Vertex ram2 = F.next(V,-1);
    util.Eiffel.assert(ram1!=null && ram2!=null);
    // create associative array between V and F of old graph and new.
    HashMap vTable = new HashMap();
    HashMap fTable = new HashMap();
    graph.Vertex[] oldVList = new graph.Vertex[this.vertexSize()];
    graph.Vertex[] newVList = new graph.Vertex[this.vertexSize()];
    Face[] oldFList = new Face[this.faceSize()];
    Face[] newFList = new Face[this.faceSize()];
    { int i=0;
    for (Enumeration E = this.vertexEnumeration(); E.hasMoreElements();/*--*/) {
      oldVList[i]=newVList[i]=(graph.Vertex) E.nextElement();
      i++;
    }}
    { int i=0;
    for (Enumeration E = this.faceEnumeration(); E.hasMoreElements();/*--*/) {
      oldFList[i]=newFList[i]= (Face) E.nextElement();
      i++;
    }}
    // create new MetricGraph
    Face[] splitFace = new Face[] {null,null};
    MetricGraph MG = new MetricGraph(this, ram1, ram2, F, splitFace,
           new graph.Vertex[0], newVList, newFList);
    util.Eiffel.assert(splitFace[0]!=null && splitFace[1]!=null);
    splitFace[0].setFinal(true);
    splitFace[1].setFinal(true);
    F.setFinal(true);
    graph.Vertex newRam1 = null;
    graph.Vertex newRam2 = null;
    util.Eiffel.assert(newVList.length==oldVList.length);
    for (int i=0;i<newVList.length;i++) {
      if (oldVList[i]==ram1)
        newRam1=newVList[i];
      if (oldVList[i]==ram2)
        newRam2=newVList[i];
    }
    util.Eiffel.assert(newRam1!=null && newRam2!=null);
    MG.checkDistinctVHash();

    MG.setMinEdge(newRam1,newRam2,minL);
    MG.setMaxEdge(newRam1,newRam2,maxL);
    // create a HashMap from new values to old.
    HashMap lookup = new HashMap();
    HashMap reverseLookup = new HashMap();
    for (int i=0;i<newVList.length;i++) {
      lookup.put(newVList[i],oldVList[i]);
      reverseLookup.put(oldVList[i],newVList[i]);
    }
    for (int i=0;i<newFList.length;i++) {
      lookup.put(newFList[i],oldFList[i]);
      reverseLookup.put(oldFList[i],newFList[i]);
    }
    // add new property of FLATQUARTER.
    for (int i=0;i<2;i++) {
      graph.Face Ftemp = splitFace[i];
      util.Eiffel.assert(Ftemp.size()==3);
      util.Eiffel.assert(Ftemp.next(newRam1,0)==newRam1);
      util.Eiffel.assert(Ftemp.next(newRam2,0)==newRam2);
      graph.Vertex Vtemp = (graph.Vertex)reverseLookup.get(V);
      util.Eiffel.assert(Vtemp!=null);
      if (Ftemp.next(Vtemp,0)!=Vtemp)
        Vtemp = (graph.Vertex)reverseLookup.get(F.next(V,2));
      util.Eiffel.assert(Ftemp.next(Vtemp,0)==Vtemp);
      util.Eiffel.assert(Vtemp!= newRam1 && Vtemp!= newRam2);
      util.Eiffel.assert(newRam1 == Ftemp.next(Vtemp,1) || newRam2 == Ftemp.next(Vtemp,1));
      util.Eiffel.assert(newRam1 == Ftemp.next(Vtemp,2) || newRam2 == Ftemp.next(Vtemp,2));
      MG.setFLATQUARTER(Vtemp,splitFace[i]);
    }
    //clone  minHeight, maxHeight, minEdge, maxEdge
    for (Enumeration E = MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      V = (graph.Vertex) E.nextElement();
      graph.Vertex Vold = (graph.Vertex) lookup.get(V);
      util.Eiffel.assert(Vold!=null);
      MG.setMinHeight(V,getMinHeight(Vold));
      MG.setMaxHeight(V,getMaxHeight(Vold));
      F = V.getAny();
      for (int i=0;i<V.size();i++) {
        graph.Vertex V1 = V.next(F,i).next(V,1);
        graph.Vertex V1old = (graph.Vertex)lookup.get(V1);
        if (isEdge(Vold,V1old)) {
          MG.setMinEdge(V,V1,getMinEdge(Vold,V1old));
          MG.setMaxEdge(V,V1,getMaxEdge(Vold,V1old));
        }
        else {
          util.Eiffel.assert(V==newRam1 || V ==newRam2);
          util.Eiffel.assert(V1==newRam1 || V1==newRam2);
        }
      }
    }
    //clone properties
    for (Iterator I = properties.iterator();I.hasNext();/*--*/) {
      Object[] obj = (Object[]) I.next();
      Object[] newKey = dict(obj,reverseLookup);
      util.Eiffel.assert(newKey!=null);
      Set S = (Set)properties.get(obj);
      for (Iterator J = S.iterator();J.hasNext();/*--*/) {
        String name = (String) J.next();
        MG.setProperty(newKey,name);
      }
    }
    return MG;
  }

  public MetricGraph fourQuarters(graph.Face F, BigDecimal minH, BigDecimal maxH) {
    util.Eiffel.assert(F.size()==4);
    util.Eiffel.assert(isORIGINALQUAD(F));
    MetricGraph MG = addUpright(F,F.getAny(),minH,maxH);
    MG = MG.moreUpright();
    util.Eiffel.assert(MG!=null);
    MG = MG.moreUpright();
    util.Eiffel.assert(MG!=null);
    util.Eiffel.assert(MG.moreUpright()==null);
    return MG;
  }

    /**
   * create a new metric graph with an upright quarter
   * The upright quarter has vertices
   * {V,F.next(V,1),new vertex};
   *
   */
  private MetricGraph addUpright(graph.Face F, graph.Vertex V,
      BigDecimal minH, BigDecimal maxH) {
    util.Eiffel.assert(this.contains(F));
    util.Eiffel.assert(this.contains(V));
    util.Eiffel.assert(F.next(V,0)==V);
    util.Eiffel.assert(F.size()==4);
    util.Eiffel.assert(F.isFinal());
    F.setFinal(false);
    graph.Vertex ram1 = V;
    graph.Vertex ram2 = F.next(V,1);
    util.Eiffel.assert(ram1!=null && ram2!=null);
    // create associative array between V and F of old graph and new.
    HashMap vTable = new HashMap();
    HashMap fTable = new HashMap();
    graph.Vertex[] oldVList = new graph.Vertex[this.vertexSize()];
    graph.Vertex[] newVList = new graph.Vertex[this.vertexSize()];
    Face[] oldFList = new Face[this.faceSize()];
    Face[] newFList = new Face[this.faceSize()];
    { int i=0;
    for (Enumeration E = this.vertexEnumeration(); E.hasMoreElements();/*--*/) {
      oldVList[i]=newVList[i]=(graph.Vertex) E.nextElement();
      i++;
    }}
    { int i=0;
    for (Enumeration E = this.faceEnumeration(); E.hasMoreElements();/*--*/) {
      oldFList[i]=newFList[i]= (Face) E.nextElement();
      i++;
    }}
    // create new MetricGraph.
    Face[] splitFace = new Face[] {null,null};
    graph.Vertex[] addedVertex = new graph.Vertex[1];
    MetricGraph MG = new MetricGraph(this, ram1, ram2, F, splitFace,
           addedVertex, newVList, newFList);
    util.Eiffel.assert(splitFace[0]!=null && splitFace[1]!=null);
    util.Eiffel.assert(splitFace[0].size()==3);
    util.Eiffel.assert(splitFace[1].size()==5);
    splitFace[0].setFinal(true);
    splitFace[1].setFinal(true);
    F.setFinal(true);
    // set min and max lengths on new edges and heights.
    graph.Vertex newRam1 = null;
    graph.Vertex newRam2 = null;
    util.Eiffel.assert(newVList.length==oldVList.length);
    for (int i=0;i<newVList.length;i++) {
      if (oldVList[i]==ram1)
        newRam1=newVList[i];
      if (oldVList[i]==ram2)
        newRam2=newVList[i];
    }
    util.Eiffel.assert(newRam1!=null && newRam2!=null);
    util.Eiffel.assert(addedVertex[0]!=null);
    MG.checkDistinctVHash();
    MG.setMinHeight(addedVertex[0],minH);
    MG.setMaxHeight(addedVertex[0],maxH);
    MG.setMinEdge(newRam1,addedVertex[0],P.I_two);
    MG.setMaxEdge(newRam1,addedVertex[0],P.eM());
    MG.setMinEdge(newRam2,addedVertex[0],P.I_two);
    MG.setMaxEdge(newRam2,addedVertex[0],P.eM());
    // create a HashMap from new values to old.
    HashMap lookup = new HashMap();
    HashMap reverseLookup = new HashMap();
    for (int i=0;i<newVList.length;i++) {
      lookup.put(newVList[i],oldVList[i]);
      reverseLookup.put(oldVList[i],newVList[i]);
    }
    for (int i=0;i<newFList.length;i++) {
      lookup.put(newFList[i],oldFList[i]);
      reverseLookup.put(oldFList[i],newFList[i]);
    }
    util.Eiffel.assert(lookup.get(splitFace[1])==F);
    // add new property of UPRIGHT.
    {
      graph.Face Ftemp = splitFace[0];
      util.Eiffel.assert(Ftemp.size()==3);
      util.Eiffel.assert(Ftemp.next(addedVertex[0],0)==addedVertex[0]);
      MG.setUPRIGHT(addedVertex[0],Ftemp);
    }
    //clone minHeight, maxHeight, minEdge, maxEdge
    for (Enumeration E = MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      V = (graph.Vertex) E.nextElement();
      if (V==addedVertex[0])
        continue;
      graph.Vertex Vold = (graph.Vertex) lookup.get(V);
      util.Eiffel.assert(Vold!=null);
      MG.setMinHeight(V,getMinHeight(Vold));
      MG.setMaxHeight(V,getMaxHeight(Vold));
      F = V.getAny();
      for (int i=0;i<V.size();i++) {
        graph.Vertex V1 = V.next(F,i).next(V,1);
        if (V1==addedVertex[0])
          continue;
        graph.Vertex V1old = (graph.Vertex)lookup.get(V1);
        util.Eiffel.assert(isEdge(Vold,V1old));
        if (isEdge(Vold,V1old)) {
          MG.setMinEdge(V,V1,getMinEdge(Vold,V1old));
          MG.setMaxEdge(V,V1,getMaxEdge(Vold,V1old));
        }
        else {
          util.Eiffel.assert(V==newRam1 || V ==newRam2);
          util.Eiffel.assert(V1==newRam1 || V1==newRam2);
        }
      }
    }
    //clone properties
    for (Iterator I = properties.iterator();I.hasNext();/*--*/) {
      Object[] obj = (Object[]) I.next();
      Object[] newKey = dict(obj,reverseLookup);
      util.Eiffel.assert(newKey!=null);
      Set S = (Set)properties.get(obj);
      for (Iterator J = S.iterator();J.hasNext();/*--*/) {
        String name = (String) J.next();
        MG.setProperty(newKey,name);
      }
    }
    return MG;
  }

  /**
   * returns null, or {V,F} of upright pair.
   */
  public Object[] getUpright() {
    graph.Face F=null;
    graph.Vertex V = null;
    // find upright quarter, whose diagonal is on a face with more than three sides.
    for (Enumeration E = this.faceEnumeration();E.hasMoreElements();/*--*/){
      graph.Face F0 = (graph.Face)E.nextElement();
      graph.Vertex W0 = F0.getAny();
      for (int i=0;i<F0.size();i++) {
        graph.Vertex W =F0.next(W0,i);
        if (isUPRIGHT(W,F0)) { // V candidate found.
          return new Object[]{W,F0};
        }
      }
    }
    return null;
  }

  /**
   * create a new metric graph with one more triangle than the current one
   * Find an upright vertex, and a face on it that is not a triangle.
   * Add there.
   * Make new lengths [2,eM()].
   * Label both as UPRIGHT.
   * returns null, if there exists none such.
   */
  private MetricGraph moreUpright() {
    graph.Face F=null;
    graph.Vertex V = null;
    util.Eiffel.assert(getUpright()!=null);
    // find upright quarter, whose diagonal is on a face with more than three sides.
    for (Enumeration E = this.faceEnumeration();E.hasMoreElements();/*--*/){
      graph.Face F0 = (graph.Face)E.nextElement();
      graph.Vertex W0 = F0.getAny();
      for (int i=0;i<F0.size();i++) {
        graph.Vertex W =F0.next(W0,i);
        if (isUPRIGHT(W,F0)) { // V candidate found.
          for (int j=0;j<W.size();j++) {
            graph.Face F1 = W.next(F0,j);
            if (F1.size()>3) {
              F = F1;
              V = W;  // found
            }
          }
        }
      }
    }
    if (F==null || V==null)
      return null;
    util.Eiffel.assert(this.contains(F));
    util.Eiffel.assert(this.contains(V));
    util.Eiffel.assert(F.next(V,0)==V);
    //util.Eiffel.assert(F.size()==4);
    util.Eiffel.assert(F.isFinal());
    F.setFinal(false);
    graph.Vertex ram1 = V;
    graph.Vertex ram2 = F.next(V,2);
    util.Eiffel.assert(ram1!=null && ram2!=null);
    // create associative array between V and F of old graph and new.
    HashMap vTable = new HashMap();
    HashMap fTable = new HashMap();
    graph.Vertex[] oldVList = new graph.Vertex[this.vertexSize()];
    graph.Vertex[] newVList = new graph.Vertex[this.vertexSize()];
    Face[] oldFList = new Face[this.faceSize()];
    Face[] newFList = new Face[this.faceSize()];
    { int i=0;
    for (Enumeration E = this.vertexEnumeration(); E.hasMoreElements();/*--*/) {
      oldVList[i]=newVList[i]=(graph.Vertex) E.nextElement();
      i++;
    }}
    { int i=0;
    for (Enumeration E = this.faceEnumeration(); E.hasMoreElements();/*--*/) {
      oldFList[i]=newFList[i]= (Face) E.nextElement();
      i++;
    }}
    // create new Metric Graph
    Face[] splitFace = new Face[] {null,null};
    MetricGraph MG = new MetricGraph(this, ram1, ram2, F, splitFace,
           new graph.Vertex[0], newVList, newFList);
    util.Eiffel.assert(splitFace[0].size()==3);
    util.Eiffel.assert(splitFace[1].size()+1==F.size());
    util.Eiffel.assert(splitFace[0]!=null && splitFace[1]!=null);
    splitFace[0].setFinal(true);
    splitFace[1].setFinal(true);
    F.setFinal(true);
    graph.Vertex newRam1 = null;
    graph.Vertex newRam2 = null;
    util.Eiffel.assert(newVList.length==oldVList.length);
    for (int i=0;i<newVList.length;i++) {
      if (oldVList[i]==ram1)
        newRam1=newVList[i];
      if (oldVList[i]==ram2)
        newRam2=newVList[i];
    }
    util.Eiffel.assert(newRam1!=null && newRam2!=null);
    MG.checkDistinctVHash();
    // set new edge Lengths.
    MG.setMinEdge(newRam1,newRam2,P.I_two);
    MG.setMaxEdge(newRam1,newRam2,P.eM());
    // create a HashMap from new values to old.
    HashMap lookup = new HashMap();
    HashMap reverseLookup = new HashMap();
    for (int i=0;i<newVList.length;i++) {
      lookup.put(newVList[i],oldVList[i]);
      reverseLookup.put(oldVList[i],newVList[i]);
    }
    for (int i=0;i<newFList.length;i++) {
      lookup.put(newFList[i],oldFList[i]);
      reverseLookup.put(oldFList[i],newFList[i]);
    }
    //set new UPRIGHT QUARTERS.
    util.Eiffel.assert(lookup.get(splitFace[1])==F);
    util.Eiffel.assert(lookup.get(newRam1)==V);
    MG.setUPRIGHT(newRam1,splitFace[0]);
    if (splitFace[1].size()==3) {
      MG.setUPRIGHT(newRam1,splitFace[1]);
    }
    //clone  minHeight, maxHeight, minEdge, maxEdge
    for (Enumeration E = MG.vertexEnumeration();E.hasMoreElements();/*--*/) {
      V = (graph.Vertex) E.nextElement();
      graph.Vertex Vold = (graph.Vertex) lookup.get(V);
      util.Eiffel.assert(Vold!=null);
      MG.setMinHeight(V,getMinHeight(Vold));
      MG.setMaxHeight(V,getMaxHeight(Vold));
      F = V.getAny();
      for (int i=0;i<V.size();i++) {
        graph.Vertex V1 = V.next(F,i).next(V,1);
        graph.Vertex V1old = (graph.Vertex)lookup.get(V1);
        if (isEdge(Vold,V1old)) {
          MG.setMinEdge(V,V1,getMinEdge(Vold,V1old));
          MG.setMaxEdge(V,V1,getMaxEdge(Vold,V1old));
        }
        else {
          util.Eiffel.assert(V==newRam1 || V ==newRam2);
          util.Eiffel.assert(V1==newRam1 || V1==newRam2);
        }
      }
    }
    //clone properties
    for (Iterator I = properties.iterator();I.hasNext();/*--*/) {
      Object[] obj = (Object[]) I.next();
      Object[] newKey = dict(obj,reverseLookup);
      util.Eiffel.assert(newKey!=null);
      Set S = (Set)properties.get(obj);
      for (Iterator J = S.iterator();J.hasNext();/*--*/) {
        String name = (String) J.next();
        MG.setProperty(newKey,name);
      }
    }
    return MG;
  }


  private MetricGraph(MetricGraph G, graph.Vertex ram1, graph.Vertex ram2,
    Face oldF, Face[] splitFace, graph.Vertex[] newVertexList, graph.Vertex[] vList, Face[] fList) {
    super(G,ram1,ram2,oldF,splitFace,newVertexList,vList,fList);
    originalFaceProfile = G.originalFaceProfile;
  }

  /**
   * This constructor is used in clone, to avoid all initialization work.
   * super(G) makes a shallow copy.
   * The dummyVar gives this constructor a different signature from MetricGraph(Graph),
   *  which is already in use as a heavily initialized constructor.
   */
  private MetricGraph(Graph G,Object dummyVar) {
    super(G);
  }

  /**
   * Shallow clone of base class Graph.
   * Deep clone of subclass data.
   */
  public Object clone() {
    MetricGraph G = new MetricGraph(this,new Object());
    G.maxEdge = (Function)this.maxEdge.clone();
    G.maxHeight = (HashMap)this.maxHeight.clone();
    G.minEdge = (Function)this.minEdge.clone();
    G.minHeight = (HashMap)this.minHeight.clone();
    for (int i=0;i<G.originalFaceProfile.length;i++) {
      G.originalFaceProfile[i]=this.originalFaceProfile[i];
    }
    G.properties = propertyClone(this.properties);
    return G;
  }

  /**
   * clones objects as well, semi-deep clone.
   * (This was the subject of an error that took me 2 days to find.)
   * (Different Metric Graph objects were working from the same property list).
   */
  private Function propertyClone(Function F) {
    Function G = new Function();
    for (Iterator I = F.iterator();I.hasNext();/*--*/) {
      Object[] key = (Object[])I.next();
      HashSet hs = (HashSet) F.get(key);
      G.add(key,hs.clone());
    }
    return G;
  }



}