package linprog;
import java.util.*;
import util.*;
import ilog.cplex.example.CpxProblem;
import java.io.*;

/**
 * Title:        KeplerConjecture 2001
 * Description:  Licensed under GNU GPL http://www.gnu.org/copyleft/gpl.html
 * Copyright:    Copyright (c) 2001, Thomas C. Hales
 *
 * @author
 * @version 2001.0
 */

/**
 * ParseIneq reads a formatted file of inequalities and converts it to an abstract syntax
 * tree.
 *
 * Template builder takes the abstract syntax tree and converts it to a list of
 * InequalityTemplate.
 *
 * The grammar of the original file is essentially built into this file.
 * It should all be better implemented with the Visitor Pattern, running through
 * the nodes of the AST.  But here we do it in a quick and dirty way.
 *
 * InequalityTemplate is used to create Inequality objects for the linear program.
 */

class TemplateBuilder {

  private Node head;
  public TemplateBuilder(Node head) {
    assertType(head,Node.INEQUALITYLIST);
    this.head = head;
  }

  // data in an InequalityTemplate;
  private HashMap VarMin = new HashMap(); // VarName->Interval
  private HashMap VarMax = new HashMap(); // VarName->Interval
  //CpxProblem.Equal CpxProblem.LessEqual CpxProblem.GreaterEqual

  private InequalityTemplate build(String name,ArrayList coeff,char sense) {
    return new InequalityTemplate(name,VarMin,VarMax,coeff,sense);
  }

  private static void assertType(Node N,int type) {
    util.Eiffel.assert(N.type()==type);
  }

 /**
  * Production rules are all essentially repeated here.
  * Should implement all this with a visitor pattern, etc.
  * But I am in a rush today, and do a quick implementation.
  */
  public ArrayList readInequalityList() {
    ArrayList ineq = new ArrayList();
    for (Iterator I = head.childIterator();I.hasNext();/*--*/) {
      Node child = (Node)I.next();
      if (child.type()==Node.DOMAIN3)
        readDomain3(child);
      else ineq.add(readInequality(child));
    }
    return ineq;
  }

  /**
   * sets instance variables VarMin and VarMax;
   */
  void readDomain3(Node N) {
    ArrayList varName = readDomainVar((Node)N.getChild(0));
    VarMin = readDomain((Node)N.getChild(1),varName);
    VarMax = readDomain((Node)N.getChild(2),varName);
  }

  private HashMap readDomain(Node N,ArrayList varName) {
    ArrayList min = new ArrayList();
    for (Iterator I = N.childIterator();I.hasNext();/*--*/) {
      Node child = (Node)I.next();
      assertType(child,Node.NUMBER);
      Interval A = (Interval) child.getObject();
      min.add(A);
    }
    util.Eiffel.assert(min.size()==varName.size(),"unmatched sizes on "+min.size()+" "+varName.size()+debugShowArrayList(varName)+
      debugShowArrayList(min));
    HashMap varmin = new HashMap();
    for (int i=0;i<min.size();i++) {
      varmin.put(varName.get(i),min.get(i));
    }
    return varmin;
  }

  private String debugShowArrayList(ArrayList A) {
    StringBuffer B = new StringBuffer();
    for (int i=0;i<A.size();i++) {
      B.append(" "+A.get(i).toString());
    }
    return B.toString();
  }

  private ArrayList readDomainVar(Node N) {
    ArrayList domainvar = new ArrayList();
    for (Iterator I = N.childIterator();I.hasNext();/*--*/){
      Node child = (Node)I.next();
      assertType(child,Node.SYMBOL);
      domainvar.add(child.getObject());
    }
    return domainvar;
  }

 private InequalityTemplate readInequality(Node N) {
  Iterator I = N.childIterator();
  //1. get Name;
  Node nameNode = (Node)I.next();
  assertType(nameNode,Node.SYMBOL);
  String name = (String)nameNode.getObject();
  //2. get LHS
  ArrayList coeffLHS = readLinCombo((Node)I.next()); // format {Interval,String,Interval,...}
  //3. get SENSE
  Node senseNode = (Node)I.next();
  assertType(senseNode,Node.SENSE);
  char c = ((String)senseNode.getObject()).charAt(0);
  char sense;
  switch (c) {
    case '<' : sense = CpxProblem.LessEqual; break;
    case '>' : sense = CpxProblem.GreaterEqual; break;
    case '=' : sense = CpxProblem.Equal; break;
    default : sense = CpxProblem.Equal;
  }
  //4. get RHS
  ArrayList coeffRHS = readLinCombo((Node)I.next());
  //5. set instance variables according to the inequality found.
  ArrayList coeff = new ArrayList();
  util.Eiffel.assert((coeffRHS.size() % 2) == 0); // check parity.
  util.Eiffel.assert((coeffLHS.size() % 2) == 0);
  for (int i=0;i< coeffLHS.size(); i+= 2) { // note unconventional increment;
    coeff.add(new Object[] {coeffLHS.get(i+1),coeffLHS.get(i)}); // { var name, interval };
  }
  // move RHS over to LHS, changing the sign in the process.
  for (int i=0;i< coeffRHS.size(); i+= 2) { // unconventional increment;
    Interval A = (Interval) coeffRHS.get(i);
    Object[] obj = new Object[] { coeffRHS.get(i+1), A.negate() };
    coeff.add(obj);
  }
  return build(name,coeff,sense);
  }

 private ArrayList readLinCombo(Node N) {
  assertType(N,Node.LINCOMBO);
  ArrayList data = new ArrayList();
  for (Iterator I=N.childIterator();I.hasNext();/*--*/) {
    Node child = (Node)I.next();
    assertType(child,Node.TERM);
    //terms have two children: an interval terminal and a symbol terminal.
    Object grandchildInterval = child.getChild(0).getObject();
    Object grandchildSymbol = child.getChild(1).getObject();
    data.add(grandchildInterval);
    data.add(grandchildSymbol);
  }
  return data;
 }

 public static class TestCode extends UnitTest {
  public static String data = ParseIneqTest.TestCode.testData;

  public void test() {
    ParseIneq P = new ParseIneq(data);
    PeekIterator I = P.Tokenize();
    Node N = P.parseInequalityList(I);
    TemplateBuilder tb = new TemplateBuilder(N);
    InequalityTemplate[] list = (InequalityTemplate[])
        tb.readInequalityList().toArray(new InequalityTemplate[0]);
    StringBuffer buff = new StringBuffer();
    for (int i=0;i<list.length;i++) {
      buff.append(list[i].toString());
    }
    String  expected =
      "# varnames: \n"+
      " [y6,y5,y4,y3,y2,y1]\n"+
      "# varmin: \n"+
      " [2`0,2`0,2`0,2`0,2`0,2`0]\n"+
      "# varmax: \n"+
      " [2.5589`0,2.5589`0,2.5589`0,2.5589`0,2.5589`0,2.5589`0]\n"+
      " +1.0`0 y1 +1.0`0 y2 +1.0`0 y3 +1.0`0 y4 +1.0`0 y5 +3.0`0 y6 -1.0`0 dih -1.0`0 unit< 0";
    assert(expected.equals(buff.substring(0,expected.length())),expected);
    }

 }

}